import sys
import traceback

from six import iteritems

from .TypicalWatchDogBase import TypicalWatchDogBase
from .JumboWatchDog import JumboWatchDog

from pandajedi.jedicore import JediCoreUtils
from pandajedi.jedicore.MsgWrapper import MsgWrapper
from pandajedi.jediconfig import jedi_config
from pandajedi.jedibrokerage import AtlasBrokerUtils

from pandaserver.taskbuffer import JobUtils
from pandaserver.dataservice import DataServiceUtils

from pandacommon.pandalogger.PandaLogger import PandaLogger

logger = PandaLogger().getLogger(__name__.split('.')[-1])

# watchdog for ATLAS production
class AtlasProdWatchDog(TypicalWatchDogBase):

    # constructor
    def __init__(self, taskBufferIF, ddmIF):
        TypicalWatchDogBase.__init__(self, taskBufferIF, ddmIF)

    # main
    def doAction(self):
        try:
            # get logger
            tmpLog = MsgWrapper(logger)
            tmpLog.debug('start')

            # action for priority boost
            self.doActionForPriorityBoost(tmpLog)

            # action for reassign
            self.doActionForReassign(tmpLog)

            # action for throttled
            self.doActionForThrottled(tmpLog)

            # action for high prio pending
            for minPriority, timeoutVal in [(950, 10),
                                            (900, 30),
                                            ]:
                self.doActionForHighPrioPending(tmpLog, minPriority, timeoutVal)

            # action to set scout job data w/o scouts
            self.doActionToSetScoutJobData(tmpLog)

            # action to throttle jobs in paused tasks
            self.doActionToThrottleJobInPausedTasks(tmpLog)

            # action for jumbo
            jumbo = JumboWatchDog(self.taskBufferIF, self.ddmIF, tmpLog, 'atlas', 'managed')
            jumbo.run()
        except Exception:
            errtype, errvalue = sys.exc_info()[:2]
            tmpLog.error('failed with {0}:{1} {2}'.format(errtype.__name__, errvalue, traceback.format_exc()))
        # return
        tmpLog.debug('done')
        return self.SC_SUCCEEDED

    # action for priority boost
    def doActionForPriorityBoost(self, gTmpLog):
        # get work queue mapper
        workQueueMapper = self.taskBufferIF.getWorkQueueMap()
        # get list of work queues
        workQueueList = workQueueMapper.getAlignedQueueList(self.vo, self.prodSourceLabel)
        resource_types = self.taskBufferIF.load_resource_types()
        # loop over all work queues
        for workQueue in workQueueList:
            break_loop = False # for special workqueues we only need to iterate once
            for resource_type in resource_types:
                gTmpLog.debug('start workQueue={0}'.format(workQueue.queue_name))
                # get tasks to be boosted
                if workQueue.is_global_share:
                    task_criteria = {'gshare': workQueue.queue_name,
                                     'resource_type': resource_type.resource_name}
                else:
                    break_loop = True
                    task_criteria = {'workQueue_ID': workQueue.queue_id}
                dataset_criteria = {'masterID': None, 'type': ['input', 'pseudo_input']}
                task_param_list = ['jediTaskID', 'taskPriority', 'currentPriority', 'parent_tid', 'gshare']
                dataset_param_list = ['nFiles', 'nFilesUsed', 'nFilesTobeUsed', 'nFilesFinished', 'nFilesFailed']
                taskVarList = self.taskBufferIF.getTasksWithCriteria_JEDI(self.vo, self.prodSourceLabel, ['running'],
                                                                          taskCriteria= task_criteria,
                                                                          datasetCriteria=dataset_criteria,
                                                                          taskParamList=task_param_list,
                                                                          datasetParamList=dataset_param_list,
                                                                          taskLockColumn='throttledTime',
                                                                          taskLockInterval=20)
                boostedPrio = 900
                toBoostRatio = 0.95
                for taskParam,datasetParam in taskVarList:
                    jediTaskID = taskParam['jediTaskID']
                    taskPriority = taskParam['taskPriority']
                    currentPriority = taskParam['currentPriority']
                    parent_tid = taskParam['parent_tid']
                    gshare = taskParam['gshare']
                    # check parent
                    parentState = None
                    if parent_tid not in [None, jediTaskID]:
                        parentState = self.taskBufferIF.checkParentTask_JEDI(parent_tid)
                        if parentState != 'completed':
                            gTmpLog.info('#ATM #KV label=managed jediTaskID={0} skip prio boost since parent_id={1} has parent_status={2}'
                                         .format(jediTaskID, parent_tid, parentState))
                            continue
                    nFiles = datasetParam['nFiles']
                    nFilesFinished = datasetParam['nFilesFinished']
                    nFilesFailed = datasetParam['nFilesFailed']
                    # get num jobs
                    nJobs = self.taskBufferIF.getNumJobsForTask_JEDI(jediTaskID)
                    nRemJobs = None
                    if nJobs is not None:
                        try:
                            nRemJobs = int(float(nFiles-nFilesFinished-nFilesFailed) * float(nJobs) / float(nFiles))
                        except Exception:
                            pass
                    tmpStr = 'jediTaskID={0} nFiles={1} nFilesFinishedFailed={2} '.format(jediTaskID,nFiles,nFilesFinished+nFilesFailed)
                    tmpStr += 'nJobs={0} nRemJobs={1} parent_tid={2} parentStatus={3}'.format(nJobs, nRemJobs, parent_tid, parentState)
                    gTmpLog.debug(tmpStr)

                    try:
                        if nRemJobs is not None and float(nFilesFinished+nFilesFailed) / float(nFiles) >= toBoostRatio and nRemJobs <= 100:
                            # skip high enough
                            if currentPriority < boostedPrio:
                                gTmpLog.info(' >>> action=priority_boosting of jediTaskID={0} to priority={1} #ATM #KV label=managed '.
                                             format(jediTaskID, boostedPrio))
                                self.taskBufferIF.changeTaskPriorityPanda(jediTaskID, boostedPrio)

                            # skip express or non global share
                            newShare = 'Express'
                            if gshare != newShare and workQueue.is_global_share:
                                gTmpLog.info(' >>> action=gshare_reassignment jediTaskID={0} from gshare_old={2} to gshare_new={1} #ATM #KV label=managed'.
                                             format(jediTaskID, newShare, gshare))
                                self.taskBufferIF.reassignShare([jediTaskID], newShare, True)
                            gTmpLog.info('>>> done jediTaskID={0}'.format(jediTaskID))
                    except Exception:
                        pass

                if break_loop:
                    break

    # action for reassignment
    def doActionForReassign(self,gTmpLog):
        # get DDM I/F
        ddmIF = self.ddmIF.getInterface(self.vo)
        # get site mapper
        siteMapper = self.taskBufferIF.getSiteMapper()
        # get tasks to get reassigned
        taskList = self.taskBufferIF.getTasksToReassign_JEDI(self.vo,self.prodSourceLabel)

        gTmpLog.debug('got {0} tasks to reassign'.format(len(taskList)))
        for taskSpec in taskList:
            tmpLog = MsgWrapper(logger, '< jediTaskID={0} >'.format(taskSpec.jediTaskID))
            tmpLog.debug('start to reassign')
            # DDM backend
            ddmBackEnd = taskSpec.getDdmBackEnd()
            # get datasets
            tmpStat,datasetSpecList = self.taskBufferIF.getDatasetsWithJediTaskID_JEDI(taskSpec.jediTaskID,['output','log'])
            if tmpStat is not True:
                tmpLog.error('failed to get datasets')
                continue
            # update DB
            if not taskSpec.useWorldCloud():
                # update cloudtasks
                tmpStat = self.taskBufferIF.setCloudTaskByUser('jedi',taskSpec.jediTaskID,taskSpec.cloud,'assigned',True)
                if tmpStat != 'SUCCEEDED':
                    tmpLog.error('failed to update CloudTasks')
                    continue
                # check cloud
                if not siteMapper.checkCloud(taskSpec.cloud):
                    tmpLog.error("cloud={0} doesn't exist".format(taskSpec.cloud))
                    continue
            else:
                # re-run task brokerage
                if taskSpec.nucleus in [None,'']:
                    taskSpec.status = 'assigning'
                    taskSpec.oldStatus = None
                    taskSpec.setToRegisterDatasets()
                    self.taskBufferIF.updateTask_JEDI(taskSpec,{'jediTaskID': taskSpec.jediTaskID},
                                                      setOldModTime=True)
                    tmpLog.debug('#ATM #KV label=managed action=trigger_new_brokerage by setting task_status={0}'.
                                 format(taskSpec.status))
                    continue

                # get nucleus
                nucleusSpec = siteMapper.getNucleus(taskSpec.nucleus)
                if nucleusSpec is None:
                    tmpLog.error("nucleus={0} doesn't exist".format(taskSpec.nucleus))
                    continue

                # set nucleus
                retMap = {taskSpec.jediTaskID: AtlasBrokerUtils.getDictToSetNucleus(nucleusSpec,datasetSpecList)}
                tmpRet = self.taskBufferIF.setCloudToTasks_JEDI(retMap)

            # get T1/nucleus
            if not taskSpec.useWorldCloud():
                t1SiteName = siteMapper.getCloud(taskSpec.cloud)['dest']
            else:
                t1SiteName = nucleusSpec.getOnePandaSite()
            t1Site = siteMapper.getSite(t1SiteName)

            # loop over all datasets
            isOK = True
            for datasetSpec in datasetSpecList:
                tmpLog.debug('dataset={0}'.format(datasetSpec.datasetName))
                if DataServiceUtils.getDistributedDestination(datasetSpec.storageToken) is not None:
                    tmpLog.debug('skip {0} is distributed'.format(datasetSpec.datasetName))
                    continue
                # get location
                location = siteMapper.getDdmEndpoint(t1Site.sitename, datasetSpec.storageToken, taskSpec.prodSourceLabel,
                                                     JobUtils.translate_tasktype_to_jobtype(taskSpec.taskType))
                # make subscription
                try:
                    tmpLog.debug('registering subscription to {0} with backend={1}'.format(location,
                                                                                           ddmBackEnd))
                    tmpStat = ddmIF.registerDatasetSubscription(datasetSpec.datasetName,location,
                                                                'Production Output',asynchronous=True)
                    if tmpStat is not True:
                        tmpLog.error("failed to make subscription")
                        isOK = False
                        break
                except Exception:
                    errtype,errvalue = sys.exc_info()[:2]
                    tmpLog.warning('failed to make subscription with {0}:{1}'.format(errtype.__name__,errvalue))
                    isOK = False
                    break
            # succeeded
            if isOK:
                # activate task
                if taskSpec.oldStatus in ['assigning','exhausted',None]:
                    taskSpec.status = 'ready'
                else:
                    taskSpec.status = taskSpec.oldStatus
                taskSpec.oldStatus = None
                self.taskBufferIF.updateTask_JEDI(taskSpec,{'jediTaskID':taskSpec.jediTaskID},
                                                  setOldModTime=True)
                tmpLog.debug('finished to reassign')

    # action for throttled tasks
    def doActionForThrottled(self,gTmpLog):
        # release tasks
        nTasks = self.taskBufferIF.releaseThrottledTasks_JEDI(self.vo,self.prodSourceLabel)
        gTmpLog.debug('released {0} tasks'.format(nTasks))

        # throttle tasks
        nTasks = self.taskBufferIF.throttleTasks_JEDI(self.vo,self.prodSourceLabel,
                                                      jedi_config.watchdog.waitForThrottled)
        gTmpLog.debug('throttled {0} tasks'.format(nTasks))

    # action for high priority pending tasks
    def doActionForHighPrioPending(self, gTmpLog, minPriority, timeoutVal):
        timeoutForPending = None
        # try to get the timeout from the config files
        if hasattr(jedi_config.watchdog,'timeoutForPendingVoLabel'):
            timeoutForPending = JediCoreUtils.getConfigParam(jedi_config.watchdog.timeoutForPendingVoLabel,self.vo,self.prodSourceLabel)
        if timeoutForPending is None:
            timeoutForPending = jedi_config.watchdog.timeoutForPending
        timeoutForPending = int(timeoutForPending) * 24
        tmpRet = self.taskBufferIF.reactivatePendingTasks_JEDI(self.vo, self.prodSourceLabel,
                                                               timeoutVal, timeoutForPending,
                                                               minPriority=minPriority)
        if tmpRet is None:
            # failed
            gTmpLog.error('failed to reactivate high priority (>{0}) tasks'.format(minPriority))
        else:
            gTmpLog.info('reactivated high priority (>{0}) {1} tasks'.format(minPriority,tmpRet))

    # action to set scout job data w/o scouts
    def doActionToSetScoutJobData(self,gTmpLog):
        tmpRet = self.taskBufferIF.setScoutJobDataToTasks_JEDI(self.vo,self.prodSourceLabel)
        if tmpRet is None:
            # failed
            gTmpLog.error('failed to set scout job data')
        else:
            gTmpLog.info('set scout job data successfully')

    # action to throttle jobs in paused tasks
    def doActionToThrottleJobInPausedTasks(self,gTmpLog):
        tmpRet = self.taskBufferIF.throttleJobsInPausedTasks_JEDI(self.vo,self.prodSourceLabel)
        if tmpRet is None:
            # failed
            gTmpLog.error('failed to thottle jobs in paused tasks')
        else:
            for jediTaskID, pandaIDs in iteritems(tmpRet):
                gTmpLog.info('throttled jobs in paused jediTaskID={0} successfully'.format(jediTaskID))
                tmpRet = self.taskBufferIF.killJobs(pandaIDs,'reassign','51',True)
                gTmpLog.info('reassigned {0} jobs in paused jediTaskID={1} with {2}'.format(len(pandaIDs), jediTaskID, tmpRet))



from pandajedi.jedicore import Interaction

# base class for task generator
class TaskGeneratorBase (object):

    def __init__(self,taskBufferIF,ddmIF):
        self.ddmIF = ddmIF
        self.taskBufferIF = taskBufferIF
        self.refresh()



    def refresh(self):
        self.siteMapper = self.taskBufferIF.getSiteMapper()



Interaction.installSC(TaskGeneratorBase)



import re
import sys

from pandajedi.jedicore.MsgWrapper import MsgWrapper
from .TypicalWatchDogBase import TypicalWatchDogBase

# logger
from pandacommon.pandalogger.PandaLogger import PandaLogger
logger = PandaLogger().getLogger(__name__.split('.')[-1])



# watchdog for general purpose
class GenWatchDog(TypicalWatchDogBase):

    # constructor
    def __init__(self, taskBufferIF, ddmIF):
        TypicalWatchDogBase.__init__(self, taskBufferIF, ddmIF)



    # main
    def doAction(self):
        # get logger
        tmpLog = MsgWrapper(logger)
        tmpLog.debug('start')
        # return
        tmpLog.debug('done')
        return self.SC_SUCCEEDED



# logger
from pandacommon.pandalogger.PandaLogger import PandaLogger
from pandajedi.jedicore.MsgWrapper import MsgWrapper

from pandajedi.jedicore.JediTaskBufferInterface import JediTaskBufferInterface
from pandajedi.jedirefine import RefinerUtils
from pandajedi.jediorder.TaskBroker import TaskBroker

from pandajedi.jediddm.DDMInterface import DDMInterface

import sys

logger = PandaLogger().getLogger('TaskBroker')
tmpLog = MsgWrapper(logger)

tbIF = JediTaskBufferInterface()
tbIF.setupInterface()

siteMapper = tbIF.getSiteMapper()

ddmIF = DDMInterface()
ddmIF.setupInterface()

jediTaskID = int(sys.argv[1])

s, taskSpec = tbIF.getTaskWithID_JEDI(jediTaskID, False)
body = TaskBroker(None, tbIF, ddmIF, taskSpec.vo, taskSpec.prodSourceLabel)
body.initializeMods(tbIF, ddmIF)

taskParam = tbIF.getTaskParamsWithID_JEDI(jediTaskID)
taskParamMap = RefinerUtils.decodeJSON(taskParam)

vo = taskParamMap['vo']
prodSourceLabel = taskParamMap['prodSourceLabel']
taskType = taskParamMap['taskType']

workQueueMapper = tbIF.getWorkQueueMap()
workQueue = workQueueMapper.getQueueWithIDGshare(taskSpec.workQueue_ID, taskSpec.gshare)

impl = body.getImpl(vo, prodSourceLabel)


tmpListItem = tbIF.getTasksToBeProcessed_JEDI(None, None, None, None, None,
                                              simTasks=[jediTaskID],
                                              readMinFiles=True)
impl.doBrokerage(tmpListItem, taskSpec.vo,
                 taskSpec.prodSourceLabel, workQueue, taskSpec.resource_type)



"""
file specification for JEDI

"""

import re
import types

from taskbuffer.FileSpec import FileSpec as JobFileSpec


class JediFileSpec(object):
    # attributes
    _attributes = ('jediTaskID','datasetID','fileID','creationDate','lastAttemptTime',
                   'lfn','GUID','type','status','fsize','checksum','scope',
                   'attemptNr','maxAttempt','nEvents','keepTrack',
                   'startEvent','endEvent','firstEvent','boundaryID','PandaID',
                   'failedAttempt','lumiBlockNr','outPandaID','maxFailure', 'ramCount',
                   'is_waiting', 'proc_status')
    # attributes which have 0 by default
    _zeroAttrs = ('fsize','attemptNr','failedAttempt','ramCount')
    # mapping between sequence and attr
    _seqAttrMap = {'fileID':'ATLAS_PANDA.JEDI_DATASET_CONT_FILEID_SEQ.nextval'}


    # constructor
    def __init__(self):
        # install attributes
        for attr in self._attributes:
            if attr in self._zeroAttrs:
                object.__setattr__(self,attr,0)
            else:
                object.__setattr__(self,attr,None)
        # map of changed attributes
        object.__setattr__(self,'_changedAttrs',{})
        # locality
        object.__setattr__(self,'locality',{})
        # source name
        object.__setattr__(self,'sourceName',None)


    # override __setattr__ to collecte the changed attributes
    def __setattr__(self,name,value):
        oldVal = getattr(self,name)
        object.__setattr__(self,name,value)
        newVal = getattr(self,name)
        # collect changed attributes
        if oldVal != newVal:
            self._changedAttrs[name] = value


    # reset changed attribute list
    def resetChangedList(self):
        object.__setattr__(self,'_changedAttrs',{})


    # return map of values
    def valuesMap(self,useSeq=False,onlyChanged=False):
        ret = {}
        for attr in self._attributes:
            # use sequence
            if useSeq and attr in self._seqAttrMap:
                continue
            # only changed attributes
            if onlyChanged:
                if attr not in self._changedAttrs:
                    continue
            val = getattr(self,attr)
            if val is None:
                if attr in self._zeroAttrs:
                    val = 0
                else:
                    val = None
            ret[':%s' % attr] = val
        return ret


    # pack tuple into FileSpec
    def pack(self,values):
        for i in range(len(self._attributes)):
            attr= self._attributes[i]
            val = values[i]
            object.__setattr__(self,attr,val)


    # return column names for INSERT
    def columnNames(cls,useSeq=False,defaultVales=None,skipDefaultAttr=False):
        if defaultVales is None:
            defaultVales = {}
        ret = ""
        for attr in cls._attributes:
            if skipDefaultAttr and (attr in cls._seqAttrMap or attr in defaultVales):
                continue
            if ret != "":
                ret += ','
            if useSeq and attr in cls._seqAttrMap:
                ret += "%s" % cls._seqAttrMap[attr]
                continue
            if attr in defaultVales:
                arg = defaultVales[attr]
                if arg is None:
                    ret += "NULL"
                elif isinstance(arg, str):
                    ret += "'{0}'".format(arg)
                else:
                    ret += "{0}".format(arg)
                continue
            ret += attr
        return ret
    columnNames = classmethod(columnNames)


    # return expression of bind variables for INSERT
    def bindValuesExpression(cls,useSeq=True):
        ret = "VALUES("
        for attr in cls._attributes:
            if useSeq and attr in cls._seqAttrMap:
                ret += "%s," % cls._seqAttrMap[attr]
            else:
                ret += ":%s," % attr
        ret = ret[:-1]
        ret += ")"
        return ret
    bindValuesExpression = classmethod(bindValuesExpression)


    # return an expression of bind variables for UPDATE to update only changed attributes
    def bindUpdateChangesExpression(self):
        ret = ""
        for attr in self._attributes:
            if attr in self._changedAttrs:
                ret += '%s=:%s,' % (attr,attr)
        ret  = ret[:-1]
        ret += ' '
        return ret


    # convert to job's FileSpec
    def convertToJobFileSpec(self,datasetSpec,setType=None,useEventService=False):
        jobFileSpec = JobFileSpec()
        jobFileSpec.fileID     = self.fileID
        jobFileSpec.datasetID  = datasetSpec.datasetID
        jobFileSpec.jediTaskID = datasetSpec.jediTaskID
        jobFileSpec.lfn        = self.lfn
        jobFileSpec.GUID       = self.GUID
        if setType is None:
            jobFileSpec.type   = self.type
        else:
            jobFileSpec.type   = setType
        jobFileSpec.scope      = self.scope
        jobFileSpec.fsize      = self.fsize
        jobFileSpec.checksum   = self.checksum
        jobFileSpec.attemptNr  = self.attemptNr
        # dataset attribute
        if datasetSpec is not None:
            # dataset
            if datasetSpec.containerName not in [None,'']:
                jobFileSpec.dataset = datasetSpec.containerName
            else:
                jobFileSpec.dataset = datasetSpec.datasetName
            if self.type in datasetSpec.getInputTypes() or setType in datasetSpec.getInputTypes():
                # prodDBlock
                jobFileSpec.prodDBlock = datasetSpec.datasetName
                # storage token
                if datasetSpec.storageToken not in ['',None]:
                    jobFileSpec.dispatchDBlockToken = datasetSpec.storageToken
            else:
                # destinationDBlock
                jobFileSpec.destinationDBlock = datasetSpec.datasetName
                # storage token
                if datasetSpec.storageToken not in ['',None]:
                    jobFileSpec.destinationDBlockToken = datasetSpec.storageToken.split('/')[0]
                # destination
                if datasetSpec.destination not in ['',None]:
                    jobFileSpec.destinationSE = datasetSpec.destination
                # set prodDBlockToken for Event Service
                if useEventService and datasetSpec.getObjectStore() is not None:
                    jobFileSpec.prodDBlockToken = 'objectstore^{0}'.format(datasetSpec.getObjectStore())
                # allow no output
                if datasetSpec.isAllowedNoOutput():
                    jobFileSpec.allowNoOutput()
        # return
        return jobFileSpec


    # convert from job's FileSpec
    def convertFromJobFileSpec(self,jobFileSpec):
        self.fileID     = jobFileSpec.fileID
        self.datasetID  = jobFileSpec.datasetID
        self.jediTaskID = jobFileSpec.jediTaskID
        self.lfn        = jobFileSpec.lfn
        self.GUID       = jobFileSpec.GUID
        self.type       = jobFileSpec.type
        self.scope      = jobFileSpec.scope
        self.fsize      = jobFileSpec.fsize
        self.checksum   = jobFileSpec.checksum
        self.attemptNr  = jobFileSpec.attemptNr
        # convert NULL to None
        for attr in self._attributes:
            val = getattr(self,attr)
            if val == 'NULL':
                object.__setattr__(self,attr,None)
        # return
        return


    # check if event-level splitting is used
    def doEventLevelSplit(self):
        if self.startEvent is None:
            return False
        return True


    # get effective number of events
    def getEffectiveNumEvents(self):
        if self.endEvent is not None and self.startEvent is not None:
            evtCounts = self.endEvent-self.startEvent+1
            if evtCounts > 0:
                return evtCounts
            return 1
        if self.nEvents is not None and self.nEvents > 0:
            return self.nEvents
        return 1


    # extract fields string
    def extractFieldsStr(self,fieldNumList):
        tmpFieldStr = ''
        try:
            tmpMidStrList = re.split('\.|_tid\d+',self.lfn)
            for tmpFieldNum in fieldNumList:
                tmpFieldStr += '.'+tmpMidStrList[tmpFieldNum-1]
        except Exception:
            pass
        return tmpFieldStr



"""
mapper to map task/job to a work queue

"""

import re

from six import iteritems

from .WorkQueue import WorkQueue, RESOURCE, ACTIVE_FUNCTIONS

class WorkQueueMapper:

    # constructor
    def __init__(self):
        # Initialize maps
        self.work_queue_map = {}
        self.work_queue_global_dic_by_name = {}
        self.work_queue_global_dic_by_id = {}

    def getSqlQuery(self):
        """
        Generates the SQL to get all work queues
        """
        sql = "SELECT {0} FROM ATLAS_PANDA.JEDI_Work_Queue".format(WorkQueue.column_names())

        return sql

    def makeMap(self, work_queues, global_leave_shares):
        """
        Creates the mapping with work queues and global shares
        :param work_queues: work queues
        :param global_leave_shares: global leave shares
        :return
        """

        # 1. add all workqueues to the map
        for wq in work_queues:
            # pack
            work_queue = WorkQueue()
            work_queue.pack(wq)

            # skip inactive queues
            if not work_queue.isActive():
                continue

            # add VO
            if work_queue.VO not in self.work_queue_map:
                self.work_queue_map[work_queue.VO] = {}

            # add type
            if work_queue.queue_type not in self.work_queue_map[work_queue.VO]:
                self.work_queue_map[work_queue.VO][work_queue.queue_type] = []

            self.work_queue_map[work_queue.VO][work_queue.queue_type].append(work_queue)
            self.work_queue_global_dic_by_name[work_queue.queue_name] = work_queue
            self.work_queue_global_dic_by_id[work_queue.queue_id] = work_queue

        # sort the queue list by order
        for vo in self.work_queue_map:
            for type in self.work_queue_map[vo]:
                # make ordered map
                ordered_map = {}
                queue_map = self.work_queue_map[vo][type]
                for wq in queue_map:
                    if wq.queue_order not in ordered_map:
                        ordered_map[wq.queue_order] = []
                    # append
                    ordered_map[wq.queue_order].append(wq)
                # make sorted list
                ordered_list = list(ordered_map.keys())
                ordered_list.sort(key=lambda x: (x is None, x))
                new_list = []
                for order_val in ordered_list:
                    new_list += ordered_map[order_val]
                # set new list
                self.work_queue_map[vo][type] = new_list

        # 2. add all the global shares
        for gs in global_leave_shares:
            work_queue_gs = WorkQueue()
            work_queue_gs.pack_gs(gs)

            if work_queue_gs.VO is None:
                vo = 'atlas'
            else:
                vo = work_queue_gs.VO

            if vo not in self.work_queue_map:
                self.work_queue_map[vo] = {}

            if work_queue_gs.queue_type not in self.work_queue_map[vo]:
                self.work_queue_map[vo][work_queue_gs.queue_type] = []

            self.work_queue_map[vo][work_queue_gs.queue_type].append(work_queue_gs)
            self.work_queue_global_dic_by_name[work_queue_gs.queue_name] = work_queue_gs
            self.work_queue_global_dic_by_id[work_queue_gs.queue_id] = work_queue_gs

        # return
        return

    def dump(self):
        """
        Creates a human-friendly string showing the work queue mappings
        :return: string representation of the work queue mappings
        """
        dump_str = 'WorkQueue mapping\n'
        for VO in self.work_queue_map:
            dump_str += '  VO=%s\n' % VO
            for type in self.work_queue_map[VO]:
                dump_str += '    type=%s\n' % type
                for workQueue in self.work_queue_map[VO]:
                    dump_str += '    %s\n' % workQueue.dump()
        # return
        return dump_str

    def getQueueWithSelParams(self, vo, type, **param_map):
        """
        Used for tagging of work queues in task refiner. Get a work queue based on the selection parameters
        :param vo: vo
        :param type: type (in practice equivalent to prodsourcelabel)
        :param param_map: parameter selection map
        :return: work queue object and explanation in case no queue was found
        """
        ret_str = ''
        if vo not in self.work_queue_map:
            ret_str = 'queues for vo=%s are undefined' % vo
        elif type not in self.work_queue_map[vo]:
            # check type
            ret_str = 'queues for type=%s are undefined in vo=%s' % (type, vo)
        else:
            for wq in self.work_queue_map[vo][type]:

                # don't return global share IDs for work queues
                if wq.is_global_share:
                    continue

                # evaluate
                try:
                    ret_queue, result = wq.evaluate(param_map)
                    if result:
                        return ret_queue, ret_str
                except Exception:
                    ret_str += '{0},'.format(wq.queue_name)

            ret_str = ret_str[:-1]
            if ret_str != '':
                new_ret_str = 'eval with VO={0} '.format(vo)
                for tmp_param_key, tmp_param_val in iteritems(param_map):
                    new_ret_str += '{0}={1} failed for {0}'.format(tmp_param_key, tmp_param_val, ret_str)
                ret_str = new_ret_str

        # no queue matched to selection parameters
        return None, ret_str

    def getQueueByName(self, vo, type, queue_name):
        """
        # get queue by name
        :param queue_name: name of the queue
        :param vo: vo
        :param type: type
        :return: queue object or None if not found
        """
        if vo in self.work_queue_map and type in self.work_queue_map[vo]:
            for wq in self.work_queue_map[vo][type]:
                if wq.queue_name == queue_name:
                    return wq
        return None

    # get queue with ID
    def getQueueWithIDGshare(self, queue_id, gshare_name):
        # 1. Check for a Resource queue
        if queue_id in self.work_queue_global_dic_by_id and self.work_queue_global_dic_by_id[queue_id].queue_function == 'Resource':
            return self.work_queue_global_dic_by_id[queue_id]

        # 2. If it wasn't a resource queue, return the global share work queue
        if gshare_name in self.work_queue_global_dic_by_name:
            return self.work_queue_global_dic_by_name[gshare_name]

        # not found
        return None

    # get queue list with VO and type
    def getAlignedQueueList(self, vo, queue_type):
        """
        NOTE: Returns ONLY resource queues and global shares (old non-resource queues are skipped)
        """
        ret_list = []

        if vo in self.work_queue_map:
            # if queue type was specified
            if queue_type not in ['', None, 'any']:
                for map_queue_type in self.work_queue_map[vo]:
                    if re.match(map_queue_type, queue_type):
                        for tmp_wq in self.work_queue_map[vo][map_queue_type]:
                            if tmp_wq.isAligned():
                                ret_list.append(tmp_wq)

            # include all queue types
            else:
                for tmp_type, tmp_wq_list in iteritems(self.work_queue_map[vo]):
                    for tmp_wq in tmp_wq_list:
                        if tmp_wq.isAligned():
                            ret_list.append(tmp_wq)

        return ret_list



import os
import sys
import copy
import re
import json
import socket
import time
import datetime
import traceback

from six import iteritems

from .WatchDogBase import WatchDogBase
from pandajedi.jediconfig import jedi_config
from pandajedi.jedicore.MsgWrapper import MsgWrapper
from pandajedi.jedicore.ThreadUtils import ListWithLock
from pandajedi.jedicore import Interaction
from pandajedi.jedibrokerage import AtlasBrokerUtils

from pandaserver.dataservice import DataServiceUtils
# from pandaserver.dataservice.Activator import Activator
from pandaserver.taskbuffer import JobUtils


# logger
from pandacommon.pandalogger.PandaLogger import PandaLogger
logger = PandaLogger().getLogger(__name__.split('.')[-1])

# dry run or not
DRY_RUN = False

# dedicated workqueue for preassigned task
magic_workqueue_id = 400
magic_workqueue_name = 'wd_queuefiller'

# wait time before reassign jobs in seconds
reassign_jobs_wait_time = 300

# queue filler watchdog for ATLAS
class AtlasQueueFillerWatchDog(WatchDogBase):

    # constructor
    def __init__(self, taskBufferIF, ddmIF):
        WatchDogBase.__init__(self, taskBufferIF, ddmIF)
        self.pid = '{0}-{1}-dog'.format(socket.getfqdn().split('.')[0], os.getpid())
        # self.cronActions = {'forPrestage': 'atlas_prs'}
        self.vo = 'atlas'
        self.prodSourceLabelList = ['managed']
        # keys for cache
        self.dc_main_key = 'AtlasQueueFillerWatchDog'
        self.dc_sub_key_pt = 'PreassignedTasks'
        self.dc_sub_key_bt = 'BlacklistedTasks'
        self.dc_sub_key_attr = 'OriginalTaskAttributes'
        self.dc_sub_key_ses = 'SiteEmptySince'
        # call refresh
        self.refresh()

    # refresh information stored in the instance
    def refresh(self):
        # work queue mapper
        self.workQueueMapper = self.taskBufferIF.getWorkQueueMap()
        # site mapper
        self.siteMapper = self.taskBufferIF.getSiteMapper()
        # all sites
        allSiteList = []
        for siteName, tmpSiteSpec in iteritems(self.siteMapper.siteSpecList):
            # if tmpSiteSpec.type == 'analysis' or tmpSiteSpec.is_grandly_unified():
            allSiteList.append(siteName)
        self.allSiteList = allSiteList

    # update preassigned task map to cache
    def _update_to_pt_cache(self, ptmap):
        data_json = json.dumps(ptmap)
        self.taskBufferIF.updateCache_JEDI(main_key=self.dc_main_key, sub_key=self.dc_sub_key_pt, data=data_json)

    # get preassigned task map from cache
    def _get_from_pt_cache(self):
        cache_spec = self.taskBufferIF.getCache_JEDI(main_key=self.dc_main_key, sub_key=self.dc_sub_key_pt)
        if cache_spec is not None:
            ret_map = json.loads(cache_spec.data)
            return ret_map
        else:
            return dict()

    # update blacklisted task map to cache
    def _update_to_bt_cache(self, btmap):
        data_json = json.dumps(btmap)
        self.taskBufferIF.updateCache_JEDI(main_key=self.dc_main_key, sub_key=self.dc_sub_key_bt, data=data_json)

    # get blacklisted task map from cache
    def _get_from_bt_cache(self):
        cache_spec = self.taskBufferIF.getCache_JEDI(main_key=self.dc_main_key, sub_key=self.dc_sub_key_bt)
        if cache_spec is not None:
            ret_map = json.loads(cache_spec.data)
            return ret_map
        else:
            return dict()

    # update task original attributes map to cache
    def _update_to_attr_cache(self, attrmap):
        data_json = json.dumps(attrmap)
        self.taskBufferIF.updateCache_JEDI(main_key=self.dc_main_key, sub_key=self.dc_sub_key_attr, data=data_json)

    # get task original attributes map from cache
    def _get_from_attr_cache(self):
        cache_spec = self.taskBufferIF.getCache_JEDI(main_key=self.dc_main_key, sub_key=self.dc_sub_key_attr)
        if cache_spec is not None:
            ret_map = json.loads(cache_spec.data)
            return ret_map
        else:
            return dict()

    # update site empty-since map to cache
    def _update_to_ses_cache(self, sesmap):
        data_json = json.dumps(sesmap)
        self.taskBufferIF.updateCache_JEDI(main_key=self.dc_main_key, sub_key=self.dc_sub_key_ses, data=data_json)

    # get site empty-since map from cache
    def _get_from_ses_cache(self):
        cache_spec = self.taskBufferIF.getCache_JEDI(main_key=self.dc_main_key, sub_key=self.dc_sub_key_ses)
        if cache_spec is not None:
            ret_map = json.loads(cache_spec.data)
            return ret_map
        else:
            return dict()

    # get process lock to preassign
    def _get_lock(self):
        return self.taskBufferIF.lockProcess_JEDI(
                vo=self.vo, prodSourceLabel='managed',
                cloud=None, workqueue_id=None, resource_name=None,
                component='AtlasQueueFillerWatchDog.preassign',
                pid=self.pid, timeLimit=5)

    # release lock
    def _release_lock(self):
        return self.taskBufferIF.unlockProcess_JEDI(
                vo=self.vo, prodSourceLabel='managed',
                cloud=None, workqueue_id=None, resource_name=None,
                component='AtlasQueueFillerWatchDog.preassign',
                pid=self.pid)

    # get map of site to list of RSEs
    def get_site_rse_map(self, prod_source_label):
        site_rse_map = {}
        for tmpPseudoSiteName in self.allSiteList:
            tmpSiteSpec = self.siteMapper.getSite(tmpPseudoSiteName)
            tmpSiteName = tmpSiteSpec.get_unified_name()
            scope_input, scope_output = DataServiceUtils.select_scope(tmpSiteSpec, prod_source_label, prod_source_label)
            try:
                endpoint_token_map = tmpSiteSpec.ddm_endpoints_input[scope_input].all
            except KeyError:
                continue
            else:
                # fill
                site_rse_map[tmpSiteName] = [ k for k, v in endpoint_token_map.items() if v.get('order_read') is not None ]
        # return
        return site_rse_map

    # get to-running rate of sites between 24 hours ago ~ 6 hours ago
    def get_site_trr_map(self):
        ret_val, ret_map = AtlasBrokerUtils.getSiteToRunRateStats(
                            self.taskBufferIF, self.vo, time_window=86400, cutoff=0, cache_lifetime=600)
        if ret_val:
            return ret_map
        else:
            return None

    # get available sites sorted list
    def get_available_sites_list(self):
        tmp_log = MsgWrapper(logger, 'get_available_sites_list')
        # initialize
        available_sites_dict = {}
        # get global share
        tmpSt, jobStatPrioMap = self.taskBufferIF.getJobStatisticsByGlobalShare(self.vo)
        if not tmpSt:
            # got nothing...
            return available_sites_dict
        # get to-running rate of sites
        site_trr_map = self.get_site_trr_map()
        if site_trr_map is None:
            return available_sites_dict
        # record for excluded site reasons
        excluded_sites_dict = {
                'not_online': set(),
                'has_minrss': set(),
                'es_jobseed': set(),
                'low_trr': set(),
                'enough_nq': set(),
            }
        # loop over sites
        for tmpPseudoSiteName in self.allSiteList:
            tmpSiteSpec = self.siteMapper.getSite(tmpPseudoSiteName)
            tmpSiteName = tmpSiteSpec.get_unified_name()
            # skip site already added
            if tmpSiteName in available_sites_dict:
                continue
            # skip site if not for production
            if not tmpSiteSpec.runs_production():
                continue
            # skip site is not online
            if tmpSiteSpec.status not in ('online'):
                excluded_sites_dict['not_online'].add(tmpSiteName)
                continue
            # skip if site has memory limitations
            if tmpSiteSpec.minrss not in (0, None):
                excluded_sites_dict['has_minrss'].add(tmpPseudoSiteName)
                continue
            # skip if site has event service jobseed
            if tmpSiteSpec.getJobSeed() in ['es']:
                excluded_sites_dict['es_jobseed'].add(tmpPseudoSiteName)
                continue
            # tmp_num_slots as  num_slots in harvester_slots
            tmp_num_slots = tmpSiteSpec.getNumStandby(None, None)
            tmp_num_slots = 0 if tmp_num_slots is None else tmp_num_slots
            # skip if site has no harvester_slots setup and has not enough activity in the past 24 hours
            site_trr = site_trr_map.get(tmpSiteName)
            if tmp_num_slots == 0 and (site_trr is None or site_trr < 0.6):
                excluded_sites_dict['low_trr'].add(tmpSiteName)
                continue
            # get nQueue and nRunning
            nRunning = AtlasBrokerUtils.getNumJobs(jobStatPrioMap, tmpSiteName, 'running')
            nQueue = 0
            for jobStatus in ['activated', 'starting']:
                nQueue += AtlasBrokerUtils.getNumJobs(jobStatPrioMap, tmpSiteName, jobStatus)
            # get nStandby; for queues that specify num_slots in harvester_slots
            tmp_core_count = tmpSiteSpec.coreCount if tmpSiteSpec.coreCount else 8
            nStandby = tmp_num_slots//tmp_core_count
            # available sites: must be idle now
            n_jobs_to_fill = max(20, max(nRunning, nStandby)*2)*0.25 - nQueue
            if n_jobs_to_fill > 0:
                available_sites_dict[tmpSiteName] = (tmpSiteName, tmpSiteSpec, n_jobs_to_fill)
            else:
                excluded_sites_dict['enough_nq'].add(tmpSiteName)
        # list
        available_sites_list = list(available_sites_dict.values())
        # sort by n_jobs_to_fill
        available_sites_list.sort(key=(lambda x: x[2]), reverse=True)
        # log message
        for reason, sites_set in excluded_sites_dict.items():
            excluded_sites_str = ','.join(sorted(sites_set))
            tmp_log.debug('excluded sites due to {reason} : {sites}'.format(reason=reason, sites=excluded_sites_str))
        included_sites_str = ','.join(sorted([ x[0] for x in available_sites_list ]))
        tmp_log.debug('included sites : {sites}'.format(reason=reason, sites=included_sites_str))
        # return
        return available_sites_list

    # get busy sites
    def get_busy_sites(self):
        busy_sites_dict = {}
        # get global share
        tmpSt, jobStatPrioMap = self.taskBufferIF.getJobStatisticsByGlobalShare(self.vo)
        if not tmpSt:
            # got nothing...
            return busy_sites_dict
        # get to-running rate of sites
        site_trr_map = self.get_site_trr_map()
        if site_trr_map is None:
            return busy_sites_dict
        # loop over sites
        for tmpPseudoSiteName in self.allSiteList:
            tmpSiteSpec = self.siteMapper.getSite(tmpPseudoSiteName)
            tmpSiteName = tmpSiteSpec.get_unified_name()
            # skip site already added
            if tmpSiteName in busy_sites_dict:
                continue
            # initialize
            is_busy = False
            # site is not online viewed as busy
            if tmpSiteSpec.status not in ('online'):
                is_busy = True
            # tmp_num_slots as  num_slots in harvester_slots
            tmp_num_slots = tmpSiteSpec.getNumStandby(None, None)
            tmp_num_slots = 0 if tmp_num_slots is None else tmp_num_slots
            # get nStandby; for queues that specify num_slots in harvester_slots
            tmp_core_count = tmpSiteSpec.coreCount if tmpSiteSpec.coreCount else 8
            nStandby = tmp_num_slots//tmp_core_count
            # get nQueue and nRunning
            nRunning = AtlasBrokerUtils.getNumJobs(jobStatPrioMap, tmpSiteName, 'running')
            nQueue = 0
            for jobStatus in ['activated', 'starting']:
                nQueue += AtlasBrokerUtils.getNumJobs(jobStatPrioMap, tmpSiteName, jobStatus)
            # busy sites
            if nQueue > max(max(20, nRunning*2)*0.375, nStandby):
                busy_sites_dict[tmpSiteName] = tmpSiteSpec
        # return
        return busy_sites_dict

    # preassign tasks to site
    def do_preassign(self):
        tmp_log = MsgWrapper(logger, 'do_preassign')
        # refresh
        self.refresh()
        # list of resource type
        resource_type_list = [ rt.resource_name for rt in self.taskBufferIF.load_resource_types() ]
        # threshold of time duration in second that the queue keeps empty to trigger preassignment
        empty_duration_threshold = 1800
        # return map
        ret_map = {
                'to_reassign': {},
            }
        # loop
        for prod_source_label in self.prodSourceLabelList:
            # site-rse map
            site_rse_map = self.get_site_rse_map(prod_source_label)
            # parameters from GDP config
            max_preassigned_tasks = self.taskBufferIF.getConfigValue(
                                        'queue_filler', 'MAX_PREASSIGNED_TASKS_{0}'.format(prod_source_label), 'jedi', self.vo)
            if max_preassigned_tasks is None:
                max_preassigned_tasks = 3
            min_files_ready = self.taskBufferIF.getConfigValue(
                                        'queue_filler', 'MIN_FILES_READY_{0}'.format(prod_source_label), 'jedi', self.vo)
            if min_files_ready is None:
                min_files_ready = 50
            min_files_remaining = self.taskBufferIF.getConfigValue(
                                        'queue_filler', 'MIN_FILES_REMAINING_{0}'.format(prod_source_label), 'jedi', self.vo)
            if min_files_remaining is None:
                min_files_remaining = 100
            # load site empty-since map from cache
            site_empty_since_map_orig = self._get_from_ses_cache()
            # available sites
            available_sites_list = self.get_available_sites_list()
            # now timestamp
            now_time = datetime.datetime.utcnow()
            now_time_ts = int(now_time.timestamp())
            # update site empty-since map
            site_empty_since_map = copy.deepcopy(site_empty_since_map_orig)
            available_site_name_list = [ x[0] for x in available_sites_list ]
            for site in site_empty_since_map_orig:
                # remove sites that are no longer empty
                if site not in available_site_name_list:
                    del site_empty_since_map[site]
            for site in available_site_name_list:
                # add newly found empty sites
                if site not in site_empty_since_map_orig:
                    site_empty_since_map[site] = now_time_ts
            self._update_to_ses_cache(site_empty_since_map)
            # evaluate sites to preaassign according to cache
            # get blacklisted_tasks_map from cache
            blacklisted_tasks_map = self._get_from_bt_cache()
            blacklisted_tasks_set = set()
            for bt_list in blacklisted_tasks_map.values():
                blacklisted_tasks_set |= set(bt_list)
            # loop over available sites to preassign
            for (site, tmpSiteSpec, n_jobs_to_fill) in available_sites_list:
                # rses of the available site
                available_rses = set()
                try:
                    available_rses.update(set(site_rse_map[site]))
                except KeyError:
                    tmp_log.debug('skipped {site} since no good RSE'.format(site=site))
                    continue
                # do not consider TAPE rses
                for rse in set(available_rses):
                    if 'TAPE' in str(rse):
                        available_rses.remove(rse)
                # skip if no rse for available site
                if not available_rses:
                    tmp_log.debug('skipped {site} since no available RSE'.format(site=site))
                    continue
                # skip if no coreCount set
                if not tmpSiteSpec.coreCount or not tmpSiteSpec.coreCount > 0:
                    tmp_log.debug('skipped {site} since coreCount is not set'.format(site=site))
                    continue
                # now timestamp
                now_time = datetime.datetime.utcnow()
                now_time_ts = int(now_time.timestamp())
                # skip if not empty for long enough
                if site not in site_empty_since_map:
                    tmp_log.error('skipped {site} since not in empty-since map (should not happen)'.format(site=site))
                    continue
                empty_duration = now_time_ts - site_empty_since_map[site]
                tmp_num_slots = tmpSiteSpec.getNumStandby(None, None)
                if empty_duration < empty_duration_threshold and not tmp_num_slots:
                    tmp_log.debug('skipped {site} since not empty for enough time ({ed}s < {edt}s)'.format(
                                    site=site, ed=empty_duration, edt=empty_duration_threshold))
                    continue
                # only simul tasks if site has fairsharePolicy setup
                processing_type_constraint = ''
                if tmpSiteSpec.fairsharePolicy not in ('NULL', None):
                    if 'type=simul:0%' in tmpSiteSpec.fairsharePolicy:
                        # skip if zero share of simul
                        tmp_log.debug('skipped {site} since with fairshare but zero for simul'.format(site=site))
                        continue
                    else:
                        processing_type_constraint = "AND t.processingType='simul' "
                # site attributes
                site_maxrss =  tmpSiteSpec.maxrss if tmpSiteSpec.maxrss not in (0, None) else 999999
                site_corecount = tmpSiteSpec.coreCount
                site_capability = str(tmpSiteSpec.capability).lower()
                # make sql parameters of rses
                available_rses = list(available_rses)
                rse_params_list = []
                rse_params_map = {}
                for j, rse in enumerate(available_rses):
                    rse_param = ':rse_{0}'.format(j + 1)
                    rse_params_list.append(rse_param)
                    rse_params_map[rse_param] = rse
                rse_params_str = ','.join(rse_params_list)
                # sql
                sql_query = (
                    "SELECT t.jediTaskID, t.workQueue_ID "
                    "FROM {jedi_schema}.JEDI_Tasks t "
                    "WHERE t.status IN ('ready','running') AND t.lockedBy IS NULL "
                        "AND t.prodSourceLabel=:prodSourceLabel "
                        "AND t.resource_type=:resource_type "
                        "AND site IS NULL "
                        "AND (COALESCE(t.baseRamCount, 0) + (CASE WHEN t.ramUnit IN ('MBPerCore','MBPerCoreFixed') THEN t.ramCount*:site_corecount ELSE t.ramCount END))*0.95 < :site_maxrss "
                        "AND t.eventService=0 "
                        "AND EXISTS ( "
                            "SELECT * FROM {jedi_schema}.JEDI_Dataset_Locality dl "
                            "WHERE dl.jediTaskID=t.jediTaskID "
                                "AND dl.rse IN ({rse_params_str}) "
                            ") "
                        "{processing_type_constraint} "
                        "AND EXISTS ( "
                            "SELECT d.datasetID FROM {jedi_schema}.JEDI_Datasets d "
                            "WHERE t.jediTaskID=d.jediTaskID AND d.type='input' "
                                "AND d.nFilesToBeUsed-d.nFilesUsed>=:min_files_ready "
                                "AND d.nFiles-d.nFilesUsed>=:min_files_remaining "
                            ") "
                    "ORDER BY t.currentPriority DESC "
                    "FOR UPDATE "
                ).format(jedi_schema=jedi_config.db.schemaJEDI,
                            rse_params_str=rse_params_str,
                            processing_type_constraint=processing_type_constraint)
                # loop over resource type
                for resource_type in resource_type_list:
                    # key name for preassigned_tasks_map = site + resource_type
                    key_name = '{0}|{1}'.format(site, resource_type)
                    # skip if not match with site capability
                    if site_capability == 'score' and not resource_type.startswith('SCORE'):
                        continue
                    elif site_capability == 'mcore' and not resource_type.startswith('MCORE'):
                        continue
                    # params map
                    params_map = {
                            ':prodSourceLabel': prod_source_label,
                            ':resource_type': resource_type,
                            ':site_maxrss': site_maxrss,
                            ':site_corecount': site_corecount,
                            ':min_files_ready': min_files_ready,
                            ':min_files_remaining': min_files_remaining,
                        }
                    params_map.update(rse_params_map)
                    # get preassigned_tasks_map from cache
                    preassigned_tasks_map = self._get_from_pt_cache()
                    preassigned_tasks_cached = preassigned_tasks_map.get(key_name, [])
                    # get task_orig_attr_map from cache
                    task_orig_attr_map = self._get_from_attr_cache()
                    # number of tasks already preassigned
                    n_preassigned_tasks = len(preassigned_tasks_cached)
                    # nuber of tasks to preassign
                    n_tasks_to_preassign = max(max_preassigned_tasks - n_preassigned_tasks, 0)
                    # preassign
                    if n_tasks_to_preassign <= 0:
                        tmp_log.debug('{key_name:<64} already has enough preassigned tasks ({n_tasks:>3}) ; skipped '.format(
                                        key_name=key_name, n_tasks=n_preassigned_tasks))
                    elif DRY_RUN:
                        dry_sql_query = (
                            "SELECT t.jediTaskID, t.workQueue_ID "
                            "FROM {jedi_schema}.JEDI_Tasks t "
                            "WHERE t.status IN ('ready','running') AND t.lockedBy IS NULL "
                                "AND t.prodSourceLabel=:prodSourceLabel "
                                "AND t.resource_type=:resource_type "
                                "AND site IS NULL "
                                "AND (COALESCE(t.baseRamCount, 0) + (CASE WHEN t.ramUnit IN ('MBPerCore','MBPerCoreFixed') THEN t.ramCount*:site_corecount ELSE t.ramCount END))*0.95 < :site_maxrss "
                                "AND t.eventService=0 "
                                "AND EXISTS ( "
                                    "SELECT * FROM {jedi_schema}.JEDI_Dataset_Locality dl "
                                    "WHERE dl.jediTaskID=t.jediTaskID "
                                        "AND dl.rse IN ({rse_params_str}) "
                                    ") "
                                "{processing_type_constraint} "
                                "AND EXISTS ( "
                                    "SELECT d.datasetID FROM {jedi_schema}.JEDI_Datasets d "
                                    "WHERE t.jediTaskID=d.jediTaskID AND d.type='input' "
                                        "AND d.nFilesToBeUsed-d.nFilesUsed>=:min_files_ready "
                                        "AND d.nFiles-d.nFilesUsed>=:min_files_remaining "
                                    ") "
                            "ORDER BY t.currentPriority DESC "
                        ).format(jedi_schema=jedi_config.db.schemaJEDI,
                                    rse_params_str=rse_params_str,
                                    processing_type_constraint=processing_type_constraint)
                        # tmp_log.debug('[dry run] {} {}'.format(dry_sql_query, params_map))
                        res = self.taskBufferIF.querySQL(dry_sql_query, params_map)
                        n_tasks = 0 if res is None else len(res)
                        if n_tasks > 0:
                            result = [ x[0] for x in res if x[0] not in preassigned_tasks_cached ]
                            updated_tasks = result[:n_tasks_to_preassign]
                            tmp_log.debug('[dry run] {key_name:<64} {n_tasks:>3} tasks would be preassigned '.format(
                                            key_name=key_name, n_tasks=n_tasks_to_preassign))
                            # update preassigned_tasks_map into cache
                            preassigned_tasks_map[key_name] = list(set(updated_tasks) | set(preassigned_tasks_cached))
                            tmp_log.debug('{} ; {}'.format(str(updated_tasks), str(preassigned_tasks_map[key_name])))
                            self._update_to_pt_cache(preassigned_tasks_map)
                    else:
                        updated_tasks_orig_attr = self.taskBufferIF.queryTasksToPreassign_JEDI(
                                                                                        sql_query, params_map, site,
                                                                                        blacklist=blacklisted_tasks_set,
                                                                                        limit=n_tasks_to_preassign)
                        if updated_tasks_orig_attr is None:
                            # dbproxy method failed
                            tmp_log.error('{key_name:<64} failed to preassign tasks '.format(
                                            key_name=key_name))
                        else:
                            n_tasks = len(updated_tasks_orig_attr)
                            if n_tasks > 0:
                                updated_tasks = [ x[0] for x in updated_tasks_orig_attr ]
                                tmp_log.info('{key_name:<64} {n_tasks:>3} tasks preassigned : {updated_tasks}'.format(
                                                key_name=key_name, n_tasks=str(n_tasks), updated_tasks=updated_tasks))
                                # update preassigned_tasks_map into cache
                                preassigned_tasks_map[key_name] = list(set(updated_tasks) | set(preassigned_tasks_cached))
                                self._update_to_pt_cache(preassigned_tasks_map)
                                # update task_orig_attr_map into cache and return map
                                for taskid, orig_attr in updated_tasks_orig_attr:
                                    taskid_str = str(taskid)
                                    task_orig_attr_map[taskid_str] = orig_attr
                                    ret_map['to_reassign'][taskid] = {
                                            'site': site,
                                            'n_jobs_to_fill': n_jobs_to_fill,
                                        }
                                self._update_to_attr_cache(task_orig_attr_map)
                                # Kibana log
                                for taskid in updated_tasks:
                                    tmp_log.debug('#ATM #KV jediTaskID={taskid} action=do_preassign site={site} rtype={rtype} preassigned '.format(
                                                    taskid=taskid, site=site, rtype=resource_type))
                            else:
                                tmp_log.debug('{key_name:<64} found no proper task to preassign'.format(key_name=key_name))
        # total preassigned tasks
        preassigned_tasks_map = self._get_from_pt_cache()
        n_pt_tot = sum([ len(pt_list) for pt_list in preassigned_tasks_map.values() ])
        tmp_log.debug('now {n_pt_tot} tasks preassigned in total'.format(n_pt_tot=n_pt_tot))
        # return
        return ret_map

    # undo preassign tasks
    def undo_preassign(self):
        tmp_log = MsgWrapper(logger, 'undo_preassign')
        # refresh
        self.refresh()
        # busy sites
        busy_sites_dict = self.get_busy_sites()
        # loop to undo preassignment
        for prod_source_label in self.prodSourceLabelList:
            # parameter from GDP config
            max_preassigned_tasks = self.taskBufferIF.getConfigValue(
                                        'queue_filler', 'MAX_PREASSIGNED_TASKS_{0}'.format(prod_source_label), 'jedi', self.vo)
            if max_preassigned_tasks is None:
                max_preassigned_tasks = 3
            min_files_ready = self.taskBufferIF.getConfigValue(
                                        'queue_filler', 'MIN_FILES_READY_{0}'.format(prod_source_label), 'jedi', self.vo)
            if min_files_ready is None:
                min_files_ready = 50
            min_files_remaining = self.taskBufferIF.getConfigValue(
                                        'queue_filler', 'MIN_FILES_REMAINING_{0}'.format(prod_source_label), 'jedi', self.vo)
            if min_files_remaining is None:
                min_files_remaining = 100
            # clean up outdated blacklist
            blacklist_duration_hours = 12
            blacklisted_tasks_map_orig = self._get_from_bt_cache()
            blacklisted_tasks_map = copy.deepcopy(blacklisted_tasks_map_orig)
            now_time = datetime.datetime.utcnow()
            min_allowed_time = now_time - datetime.timedelta(hours=blacklist_duration_hours)
            min_allowed_ts = int(min_allowed_time.timestamp())
            for ts_str in blacklisted_tasks_map_orig:
                ts = int(ts_str)
                if ts < min_allowed_ts:
                    del blacklisted_tasks_map[ts_str]
            self._update_to_bt_cache(blacklisted_tasks_map)
            n_bt_old = sum([ len(bt_list) for bt_list in blacklisted_tasks_map_orig.values() ])
            n_bt = sum([ len(bt_list) for bt_list in blacklisted_tasks_map.values() ])
            tmp_log.debug('done cleanup blacklist; before {n_bt_old} , now {n_bt} tasks in blacklist'.format(n_bt_old=n_bt_old, n_bt=n_bt))
            # get a copy of preassigned_tasks_map from cache
            preassigned_tasks_map_orig = self._get_from_pt_cache()
            preassigned_tasks_map = copy.deepcopy(preassigned_tasks_map_orig)
            # clean up task_orig_attr_map in cache
            task_orig_attr_map_orig = self._get_from_attr_cache()
            task_orig_attr_map = copy.deepcopy(task_orig_attr_map_orig)
            all_preassiged_taskids = set()
            for taskid_list in preassigned_tasks_map_orig.values():
                all_preassiged_taskids |= set(taskid_list)
            for taskid_str in task_orig_attr_map_orig:
                taskid = int(taskid_str)
                if taskid not in all_preassiged_taskids:
                    del task_orig_attr_map[taskid_str]
            self._update_to_attr_cache(task_orig_attr_map)
            # loop on preassigned tasks in cache
            for key_name in preassigned_tasks_map_orig:
                # parse key name = site + resource_type
                site, resource_type = key_name.split('|')
                # preassigned tasks in cache
                preassigned_tasks_cached = preassigned_tasks_map.get(key_name, [])
                # force_undo=True for all tasks in busy sites, and force_undo=False for tasks not in status to generate jobs
                force_undo = False
                if site in busy_sites_dict or len(preassigned_tasks_cached) > max_preassigned_tasks:
                    force_undo = True
                reason_str = 'site busy or offline or with too many preassigned tasks' if force_undo \
                                else 'task paused/terminated or without enough files to process'
                # parameters for undo, kinda ugly
                params_map = {
                        ':min_files_ready': min_files_ready,
                        ':min_files_remaining': min_files_remaining,
                    }
                # undo preassign
                had_undo = False
                updated_tasks = []
                if DRY_RUN:
                    if force_undo:
                        updated_tasks = list(preassigned_tasks_cached)
                        n_tasks = len(updated_tasks)
                    else:
                        preassigned_tasks_list = []
                        preassigned_tasks_params_map = {}
                        for j, taskid in enumerate(preassigned_tasks_cached):
                            pt_param = ':pt_{0}'.format(j + 1)
                            preassigned_tasks_list.append(pt_param)
                            preassigned_tasks_params_map[pt_param] = taskid
                        if not preassigned_tasks_list:
                            continue
                        preassigned_tasks_params_str = ','.join(preassigned_tasks_list)
                        dry_sql_query = (
                            "SELECT t.jediTaskID "
                            "FROM {jedi_schema}.JEDI_Tasks t "
                            "WHERE t.jediTaskID IN ({preassigned_tasks_params_str}) "
                                "AND t.site IS NOT NULL "
                                "AND NOT ( "
                                        "t.status IN ('ready','running') "
                                        "AND EXISTS ( "
                                            "SELECT d.datasetID FROM {0}.JEDI_Datasets d "
                                            "WHERE t.jediTaskID=d.jediTaskID AND d.type='input' "
                                                "AND d.nFilesToBeUsed-d.nFilesUsed>=:min_files_ready AND d.nFiles-d.nFilesUsed>=:min_files_remaining "
                                            ") "
                                    ") "
                        ).format(jedi_schema=jedi_config.db.schemaJEDI, preassigned_tasks_params_str=preassigned_tasks_params_str)
                        res = self.taskBufferIF.querySQL(dry_sql_query, preassigned_tasks_params_map)
                        n_tasks = 0 if res is None else len(res)
                        if n_tasks > 0:
                            updated_tasks = [ x[0] for x in res ]
                    # tmp_log.debug('[dry run] {} {} force={}'.format(key_name, str(updated_tasks), force_undo))
                    had_undo = True
                    if n_tasks > 0:
                        tmp_log.debug('[dry run] {key_name:<64} {n_tasks:>3} preassigned tasks would be undone ({reason_str}) '.format(
                                        key_name=key_name, n_tasks=n_tasks, reason_str=reason_str))
                else:
                    updated_tasks = self.taskBufferIF.undoPreassignedTasks_JEDI(preassigned_tasks_cached,
                                                                                task_orig_attr_map=task_orig_attr_map,
                                                                                params_map=params_map,
                                                                                force=force_undo)
                    if updated_tasks is None:
                        # dbproxy method failed
                        tmp_log.error('{key_name:<64} failed to undo preassigned tasks (force={force_undo})'.format(
                                        key_name=key_name, force_undo=force_undo))
                    else:
                        had_undo = True
                        n_tasks = len(updated_tasks)
                        if n_tasks > 0:
                            tmp_log.info('{key_name:<64} {n_tasks:>3} preassigned tasks undone ({reason_str}) : {updated_tasks} '.format(
                                            key_name=key_name, n_tasks=str(n_tasks), reason_str=reason_str, updated_tasks=updated_tasks))
                            # Kibana log
                            for taskid in updated_tasks:
                                tmp_log.debug('#ATM #KV jediTaskID={taskid} action=undo_preassign site={site} rtype={rtype} un-preassinged since {reason_str}'.format(
                                                taskid=taskid, site=site, rtype=resource_type, reason_str=reason_str))
                # update preassigned_tasks_map into cache
                if had_undo:
                    if force_undo:
                        del preassigned_tasks_map[key_name]
                    else:
                        tmp_tasks_set = set(preassigned_tasks_cached) - set(updated_tasks)
                        if not tmp_tasks_set:
                            del preassigned_tasks_map[key_name]
                        else:
                            preassigned_tasks_map[key_name] = list(tmp_tasks_set)
                    self._update_to_pt_cache(preassigned_tasks_map)
                # update blacklisted_tasks_map into cache
                if had_undo and not force_undo:
                    blacklisted_tasks_map_orig = self._get_from_bt_cache()
                    blacklisted_tasks_map = copy.deepcopy(blacklisted_tasks_map_orig)
                    now_time = datetime.datetime.utcnow()
                    now_rounded_ts = int(now_time.replace(minute=0, second=0, microsecond=0).timestamp())
                    ts_str = str(now_rounded_ts)
                    if ts_str in blacklisted_tasks_map_orig:
                        tmp_bt_list = blacklisted_tasks_map[ts_str]
                        blacklisted_tasks_map[ts_str] = list(set(tmp_bt_list)|set(updated_tasks))
                    else:
                        blacklisted_tasks_map[ts_str] = list(updated_tasks)
                    self._update_to_bt_cache(blacklisted_tasks_map)

    # close and reassign jobs of preassigned tasks
    def reassign_jobs(self, to_reassign_map):
        tmp_log = MsgWrapper(logger, 'reassign_jobs')
        for jedi_taskid, value_map in to_reassign_map.items():
            site = value_map['site']
            n_jobs_to_fill = value_map['n_jobs_to_fill']
            # compute n_jobs_to_close from n_jobs_to_fill
            n_jobs_to_close = int(n_jobs_to_fill/3)
            # reassign
            n_jobs_closed = self.taskBufferIF.reassignJobsInPreassignedTask_JEDI(jedi_taskid, site, n_jobs_to_close)
            if n_jobs_closed is None:
                tmp_log.debug('jediTaskID={0} no longer ready/running or not assigned to {1} , skipped'.format(jedi_taskid, site))
            else:
                tmp_log.debug('jediTaskID={0} to {1} , closed {2} jobs'.format(jedi_taskid, site, n_jobs_closed))

    # main
    def doAction(self):
        try:
            # get logger
            origTmpLog = MsgWrapper(logger)
            origTmpLog.debug('start')
            # lock
            got_lock = self._get_lock()
            if not got_lock:
                origTmpLog.debug('locked by another process. Skipped')
                return self.SC_SUCCEEDED
            origTmpLog.debug('got lock')
            # undo preassigned tasks
            self.undo_preassign()
            # preassign tasks to sites
            ret_map = self.do_preassign()
            # unlock
            # self._release_lock()
            # origTmpLog.debug('released lock')
            # to-reassign map
            to_reassign_map = ret_map['to_reassign']
            if to_reassign_map:
                # wait some minutes so that preassigned tasks can be brokered, before reassigning jobs
                origTmpLog.debug('wait {0}s before reassigning jobs'.format(reassign_jobs_wait_time))
                time.sleep(reassign_jobs_wait_time)
                # reassign jobs of preassigned tasks
                self.reassign_jobs(to_reassign_map)
        except Exception:
            errtype, errvalue = sys.exc_info()[:2]
            err_str = traceback.format_exc()
            origTmpLog.error('failed with {0} {1} ; {2}'.format(errtype, errvalue, err_str))
        # return
        origTmpLog.debug('done')
        return self.SC_SUCCEEDED



from pandajedi.jedicore.Interaction import CommandReceiveInterface

# base class to interact with DDM
class DDMClientBase(CommandReceiveInterface):

    # constructor
    def __init__(self,con):
        CommandReceiveInterface.__init__(self,con)



    # list dataset/container
    def listDatasets(self,datasetName,ignorePandaDS=True):
        return self.SC_SUCCEEDED,[datasetName]



import json

from pandajedi.jedimsgprocessor.base_msg_processor import BaseMsgProcPlugin

from pandacommon.pandalogger import logger_utils


# logger
base_logger = logger_utils.setup_logger(__name__.split('.')[-1])


# processing message processing plugin
class ProcessingMsgProcPlugin(BaseMsgProcPlugin):

    def process(self, msg_obj, decoded_data=None):
        # logger
        tmp_log = logger_utils.make_logger(base_logger, method_name='process')
        # start
        tmp_log.info('start')
        tmp_log.debug('sub_id={0} ; msg_id={1}'.format(msg_obj.sub_id, msg_obj.msg_id))
        # parse
        if decoded_data is None:
            # json decode
            try:
                msg_dict = json.loads(msg_obj.data)
            except Exception as e:
                err_str = 'failed to parse message json {2} , skipped. {0} : {1}'.format(e.__class__.__name__, e, msg_obj.data)
                tmp_log.error(err_str)
                raise
        else:
            msg_dict = decoded_data
        # sanity check
        try:
            jeditaskid = int(msg_dict['workload_id'])
            # message type
            msg_type = msg_dict['msg_type']
            if msg_type == 'file_processing':
                target_list = msg_dict['files']
            elif msg_type == 'collection_processing':
                target_list = msg_dict['collections']
            elif msg_type == 'work_processing':
                pass
            else:
                raise ValueError('invalid msg_type value: {0}'.format(msg_type))
            # relation type
            relation_type = msg_dict.get('relation_type')
        except Exception as e:
            err_str = 'failed to parse message object dict {2} , skipped. {0} : {1}'.format(e.__class__.__name__, e, msg_dict)
            tmp_log.error(err_str)
            raise
        # run
        try:
            # initialize to_proceed
            to_proceed = False
            # type filters
            if msg_type in ['file_processing', 'collection_processing'] \
                    and relation_type in ['input']:
                to_proceed = True
            # whether to proceed the targets
            if to_proceed:
                # initialize
                scope_name_list_map = {}
                missing_files_list = []
                # loop over targets
                for target in target_list:
                    name = target['name']
                    scope = target['scope']
                    if (msg_type == 'file_processing' and target['status'] in ['Available']) \
                            or (msg_type == 'collection_processing' and target['status'] in ['Closed']):
                        scope_name_list_map.setdefault(scope, [])
                        scope_name_list_map[scope].append(name)
                    elif (msg_type == 'file_processing' and target['status'] in ['Missing']):
                        # missing files
                        missing_files_list.append(name)
                    else:
                        # got target in bad attributes, do nothing
                        tmp_log.debug('jeditaskid={jeditaskid}, scope={scope}, msg_type={msg_type}, status={status}, did nothing for bad target'.format(
                                        jeditaskid=jeditaskid, scope=scope, msg_type=msg_type, status=target['status']))
                        pass
                # run by each scope
                for scope, name_list in scope_name_list_map.items():
                    # about files or datasets in good status
                    if msg_type == 'file_processing':
                        tmp_log.debug('jeditaskid={0}, scope={1}, update about files...'.format(jeditaskid, scope))
                        res = self.tbIF.updateInputFilesStagedAboutIdds_JEDI(jeditaskid, scope, name_list)
                        if res is None:
                            # got error and rollback in dbproxy
                            err_str = 'jeditaskid={0}, scope={1}, failed to update files'.format(jeditaskid, scope)
                            raise RuntimeError(err_str)
                        tmp_log.info('jeditaskid={0}, scope={1}, updated {2} files'.format(jeditaskid, scope, res))
                    elif msg_type == 'collection_processing':
                        tmp_log.debug('jeditaskid={0}, scope={1}, update about datasets...'.format(jeditaskid, scope))
                        res = self.tbIF.updateInputDatasetsStagedAboutIdds_JEDI(jeditaskid, scope, name_list)
                        if res is None:
                            # got error and rollback in dbproxy
                            err_str = 'jeditaskid={0}, scope={1}, failed to update datasets'.format(jeditaskid, scope)
                            raise RuntimeError(err_str)
                        tmp_log.info('jeditaskid={0}, scope={1}, updated {2} datasets'.format(jeditaskid, scope, res))
                    # check if all ok
                    if res == len(target_list):
                        tmp_log.debug('jeditaskid={0}, scope={1}, all OK'.format(jeditaskid, scope))
                    elif res < len(target_list):
                        tmp_log.warning('jeditaskid={0}, scope={1}, only {2} out of {3} done...'.format(
                                                                                jeditaskid, scope, res, len(target_list)))
                    elif res > len(target_list):
                        tmp_log.warning('jeditaskid={0}, scope={1}, strangely, {2} out of {3} done...'.format(
                                                                                jeditaskid, scope, res, len(target_list)))
                    else:
                        tmp_log.warning('jeditaskid={0}, scope={1}, something unwanted happened...'.format(
                                                                                jeditaskid, scope))
                # handle missing files
                n_missing = len(missing_files_list)
                if n_missing > 0:
                    res = self.tbIF.setMissingFilesAboutIdds_JEDI(jeditaskid=jeditaskid, filenames=missing_files_list)
                    if res == n_missing:
                        tmp_log.debug('jeditaskid={0}, marked all {1} files missing'.format(jeditaskid, n_missing))
                    elif res < n_missing:
                        tmp_log.warning('jeditaskid={0}, only {1} out of {2} files marked missing...'.format(
                                                                                jeditaskid, res, n_missing))
                    elif res > n_missing:
                        tmp_log.warning('jeditaskid={0}, strangely, {1} out of {2} files marked missing...'.format(
                                                                                jeditaskid, res, n_missing))
                    else:
                        tmp_log.warning('jeditaskid={0}, res={1}, something unwanted happened about missing files...'.format(
                                                                                jeditaskid, res))
            else:
                # do nothing
                tmp_log.debug('jeditaskid={jeditaskid}, msg_type={msg_type}, relation_type={relation_type}, nothing done'.format(
                                jeditaskid=jeditaskid, msg_type=msg_type, relation_type=relation_type))
        except Exception as e:
            err_str = 'failed to process the message, skipped. {0} : {1}'.format(e.__class__.__name__, e)
            tmp_log.error(err_str)
            raise
        # done
        tmp_log.info('done')



import sys
import time
import random

from pandajedi.jedicore import Interaction
from pandajedi.jedicore.ThreadUtils import ZombiCleaner


class JediKnight(Interaction.CommandReceiveInterface):
    # constructor
    def __init__(self, commuChannel, taskBufferIF, ddmIF, logger, **kwargs):
        Interaction.CommandReceiveInterface.__init__(self, commuChannel)
        self.taskBufferIF = taskBufferIF
        self.ddmIF = ddmIF
        self.logger = logger
        # intra-node message broker proxies
        self.mb_proxy_dict = kwargs.get('mb_proxy_dict')
        # start zombie cleaner
        ZombiCleaner().start()

    # start communication channel in a thread
    def start(self):
        # start communication channel
        import threading
        thr = threading.Thread(target=self.startImpl)
        thr.start()

    # implementation of start()
    def startImpl(self):
        try:
            Interaction.CommandReceiveInterface.start(self)
        except Exception:
            errtype, errvalue = sys.exc_info()[:2]
            self.logger.error('crashed in JediKnight.startImpl() with %s %s' % (errtype.__name__, errvalue))

    # parse init params
    def parseInit(self, par):
        if isinstance(par, list):
            return par
        try:
            return par.split('|')
        except Exception:
            return [par]

    # sleep to avoid synchronization of loop
    def randomSleep(self, min_val=0, default_max_val=30, max_val=None):
        if max_val is None:
            max_val = default_max_val
        max_val = min(max_val, default_max_val)
        time.sleep(random.randint(min_val, max_val))


# install SCs
Interaction.installSC(JediKnight)



import os
import re
import sys
import json

from six import iteritems

from liveconfigparser.LiveConfigParser import LiveConfigParser

# get ConfigParser
tmpConf = LiveConfigParser()

# read
tmpConf.read('panda_jedi.cfg')

# dummy section class
class _SectionClass:
    pass

# load configmap
config_map_data = {}
if 'PANDA_HOME' in os.environ:
    config_map_name = 'panda_jedi_config.json'
    config_map_path = os.path.join(os.environ['PANDA_HOME'], 'etc/config_json', config_map_name)
    if os.path.exists(config_map_path):
        with open(config_map_path) as f:
            config_map_data = json.load(f)

# loop over all sections
for tmpSection in tmpConf.sections():
    # read section
    tmpDict = getattr(tmpConf, tmpSection)
    # load configmap
    if tmpSection in config_map_data:
        tmpDict.update(config_map_data[tmpSection])
    # make section class
    tmpSelf = _SectionClass()
    # update module dict
    sys.modules[ __name__ ].__dict__[tmpSection] = tmpSelf
    # expand all values
    for tmpKey,tmpVal in iteritems(tmpDict):
        # convert string to bool/int
        if tmpVal == 'True':
            tmpVal = True
        elif tmpVal == 'False':
            tmpVal = False
        elif tmpVal == 'None':
            tmpVal = None
        elif isinstance(tmpVal, str) and re.match('^\d+$',tmpVal):
            tmpVal = int(tmpVal)
        # update dict
        setattr(tmpSelf,tmpKey,tmpVal)
    # database info via env
    if tmpSection == 'db':
        if 'PANDA_DB_HOST' in os.environ:
            tmpSelf.dbhost = os.environ['PANDA_DB_HOST']
        if 'PANDA_DB_PASSWORD' in os.environ:
            tmpSelf.dbpasswd = os.environ['PANDA_DB_PASSWORD']


import os
import sys
import time
import types
import signal
import datetime
import multiprocessing
import multiprocessing.reduction

try:
    from multiprocessing.connection import reduce_connection
except ImportError:
    from multiprocessing.reduction import reduce_connection

from six import iteritems

# patch multiprocessing
from . import JediPatch

#import multiprocessing
#logger = multiprocessing.log_to_stderr()
#logger.setLevel(multiprocessing.SUBDEBUG)

###########################################################
#
# status codes for IPC
#

# class for status code
class StatusCode(object):
    def __init__(self,value):
        self.value = value

    def __str__(self):
        return "%s" % self.value

    # comparator
    def __eq__(self,other):
        try:
            return self.value == other.value
        except Exception:
            return False

    def __ne__(self,other):
        try:
            return self.value != other.value
        except Exception:
            return True



# mapping to accessors
statusCodeMap = {'SC_SUCCEEDED': StatusCode(0),
                 'SC_FAILED'   : StatusCode(1),
                 'SC_FATAL'    : StatusCode(2),
                 }


# install the list of status codes to a class
def installSC(cls):
    for sc,val in iteritems(statusCodeMap):
        setattr(cls,sc,val)


# install SCs in this module
installSC(sys.modules[ __name__ ])


###########################################################
#
# classes for IPC
#

# log message with timestamp
def dumpStdOut(sender,message):
    timeNow = datetime.datetime.utcnow()
    print("{0} {1}: INFO    {2}".format(str(timeNow),sender,message))



# object class for command
class CommandObject(object):

    # constructor
    def __init__(self,methodName,argList,argMap):
        self.methodName = methodName
        self.argList = argList
        self.argMap = argMap



# object class for response
class ReturnObject(object):

    # constructor
    def __init__(self):
        self.statusCode  = None
        self.errorValue  = None
        self.returnValue = None



# process class
class ProcessClass(object):
    # constructor
    def __init__(self,pid,connection):
        self.pid = pid
        self.nused = 0
        self.usedMemory = 0
        self.nMemLookup = 20
        # reduce connection to make it picklable
        self.reduced_pipe = reduce_connection(connection)

    # get connection
    def connection(self):
        # rebuild connection
        return self.reduced_pipe[0](*self.reduced_pipe[1])

    # reduce connection
    def reduceConnection(self,connection):
        self.reduced_pipe = reduce_connection(connection)

    # get memory usage
    def getMemUsage(self):
        # update memory info
        if self.nused % self.nMemLookup == 0:
            try:
                # read memory info from /proc
                t = open('/proc/{0}/status'.format(self.pid))
                v = t.read()
                t.close()
                value = 0
                for line in v.split('\n'):
                    if line.startswith('VmRSS'):
                        items = line.split()
                        value = int(items[1])
                        if items[2] in ['kB','KB']:
                            value /= 1024
                        elif items[2] in ['mB','MB']:
                            pass
                        break
                self.usedMemory = value
            except Exception:
                pass
            return self.usedMemory
        else:
            return None


# method class
class MethodClass(object):
    # constructor
    def __init__(self,className,methodName,vo,connectionQueue,voIF):
        self.className = className
        self.methodName = methodName
        self.vo = vo
        self.connectionQueue = connectionQueue
        self.voIF = voIF
        self.pipeList = []

    # method emulation
    def __call__(self,*args,**kwargs):
        commandObj = CommandObject(self.methodName,
                                   args,kwargs)
        nTry = 3
        for iTry in range(nTry):
            # exceptions
            retException = None
            strException = None
            try:
                stepIdx = 0
                # get child process
                child_process = self.connectionQueue.get()
                # get pipe
                stepIdx = 1
                pipe = child_process.connection()
                # get ack
                stepIdx = 2
                timeoutPeriodACK = 30
                if not pipe.poll(timeoutPeriodACK):
                    raise JEDITimeoutError("did not get ACK for %ssec" % timeoutPeriodACK)
                ack = pipe.recv()
                # send command
                stepIdx = 3
                pipe.send(commandObj)
                # wait response
                stepIdx = 4
                timeoutPeriod = 600
                timeNow = datetime.datetime.utcnow()
                if not pipe.poll(timeoutPeriod):
                    raise JEDITimeoutError("did not get response for %ssec" % timeoutPeriod)
                regTime = datetime.datetime.utcnow() - timeNow
                if regTime > datetime.timedelta(seconds=60):
                    dumpStdOut(self.className,
                              'methodName={} took {}.{:03d} sec in pid={}'.format(self.methodName,
                                                                                  regTime.seconds,
                                                                                  int(regTime.microseconds / 1000),
                                                                                  child_process.pid))
                # get response
                stepIdx = 5
                ret = pipe.recv()
                # set exception type based on error
                stepIdx = 6
                if ret.statusCode == SC_FAILED:
                    retException = JEDITemporaryError
                elif ret.statusCode == SC_FATAL:
                    retException = JEDIFatalError
            except Exception:
                errtype,errvalue = sys.exc_info()[:2]
                retException = errtype
                argStr = 'args=%s kargs=%s' % (str(args),str(kwargs))
                strException = 'VO=%s type=%s stepIdx=%s : %s.%s %s %s' % \
                               (self.vo,errtype.__name__,stepIdx,
                                self.className,self.methodName,errvalue,
                                argStr[:200])
            # increment nused
            child_process.nused += 1
            # memory check
            largeMemory = False
            memUsed = child_process.getMemUsage()
            if memUsed is not None:
                memStr= 'pid={0} memory={1}MB'.format(child_process.pid,memUsed)
                if memUsed > 1.5*1024:
                    largeMemory = True
                    memStr += ' exceeds memory limit'
                    dumpStdOut(self.className,memStr)
            # kill old or problematic process
            if child_process.nused > 1000 or retException not in [None,JEDITemporaryError,JEDIFatalError] or \
                    largeMemory:
                dumpStdOut(self.className,'methodName={0} ret={1} nused={2} {3} in pid={4}'.format(self.methodName,retException,
                                                                                                   child_process.nused,
                                                                                                   strException,
                                                                                                   child_process.pid))
                # close connection
                try:
                    pipe.close()
                except Exception:
                    pass
                # terminate child process
                try:
                    dumpStdOut(self.className,'killing pid={0}'.format(child_process.pid))
                    os.kill(child_process.pid,signal.SIGKILL)
                    dumpStdOut(self.className,'waiting pid={0}'.format(child_process.pid))
                    os.waitpid(child_process.pid,0)
                    dumpStdOut(self.className,'terminated pid={0}'.format(child_process.pid))
                except Exception:
                    errtype,errvalue = sys.exc_info()[:2]
                    if 'No child processes' not in str(errvalue):
                        dumpStdOut(self.className,'failed to terminate {0} with {1}:{2}'.format(child_process.pid,
                                                                                                errtype,errvalue))
                # make new child process
                self.voIF.launchChild()
            else:
                # reduce process object to avoid deadlock due to rebuilding of connection
                child_process.reduceConnection(pipe)
                self.connectionQueue.put(child_process)
            # success, fatal error, or maximally attempted
            if retException in [None,JEDIFatalError] or (iTry+1 == nTry):
                break
            # sleep
            time.sleep(1)
        # raise exception
        if retException is not None:
            if strException is None:
                strException = 'VO={0} {1}'.format(self.vo,ret.errorValue)
            raise retException(strException)
        # return
        if ret.statusCode == SC_SUCCEEDED:
            return ret.returnValue
        else:
            raise retException('VO=%s %s' % (self.vo,ret.errorValue))



# interface class to send command
class CommandSendInterface(object):
    # constructor
    def __init__(self,vo,maxChild,moduleName,className):
        self.vo = vo
        self.maxChild = maxChild
        self.connectionQueue = multiprocessing.Queue(maxChild)
        self.moduleName = moduleName
        self.className  = className


    # factory method
    def __getattr__(self,attrName):
        return MethodClass(self.className,attrName,self.vo,self.connectionQueue,self)


    # launcher for child processe
    def launcher(self,channel):
        # import module
        mod = __import__(self.moduleName)
        for subModuleName in self.moduleName.split('.')[1:]:
            mod = getattr(mod,subModuleName)
        # get class
        cls = getattr(mod,self.className)
        # start child process
        msg = "start {0} with pid={1}".format(self.className,os.getpid())
        dumpStdOut(self.moduleName,msg)
        timeNow = datetime.datetime.utcnow()
        try:
            cls(channel).start()
        except Exception:
            errtype,errvalue = sys.exc_info()[:2]
            dumpStdOut(self.className,'launcher crashed with {0}:{1}'.format(errtype,errvalue))



    # launch child processes to interact with DDM
    def launchChild(self):
        # make pipe
        parent_conn, child_conn = multiprocessing.Pipe()
        # make child process
        child_process = multiprocessing.Process(target=self.launcher,
                                                args=(child_conn,))
        # start child process
        child_process.daemon = True
        child_process.start()
        # keep process in queue
        processObj = ProcessClass(child_process.pid,parent_conn)
        self.connectionQueue.put(processObj)


    # initialize
    def initialize(self):
        for i in range(self.maxChild):
            self.launchChild()



# interface class to receive command
class CommandReceiveInterface(object):

    # constructor
    def __init__(self,con):
        self.con = con
        self.cacheMap = {}


    # make key for cache
    def makeKey(self,className,methodName,argList,argMap):
        try:
            tmpKey = '{0}:{1}:'.format(className,methodName)
            for argItem in argList:
                tmpKey += '{0}:'.format(str(argItem))
            for argKey,argVal in iteritems(argMap):
                tmpKey += '{0}={1}:'.format(argKey,str(argVal))
            tmpKey = tmpKey[:-1]
            return tmpKey
        except Exception:
            return None


    # main loop
    def start(self):
        while True:
            # send ACK
            self.con.send('ack')
            # get command
            commandObj = self.con.recv()
            # make return
            retObj = ReturnObject()
            # get class name
            className = self.__class__.__name__
            # check method name
            if not hasattr(self,commandObj.methodName):
                # method not found
                retObj.statusCode = self.SC_FATAL
                retObj.errorValue = 'type=AttributeError : %s instance has no attribute %s' % \
                    (className,commandObj.methodName)
            else:
                try:
                    # use cache
                    useCache = False
                    doExec = True
                    if 'useResultCache' in commandObj.argMap:
                        # get time range
                        timeRange = commandObj.argMap['useResultCache']
                        # delete from args map
                        del commandObj.argMap['useResultCache']
                        # make key for cache
                        tmpCacheKey = self.makeKey(className,commandObj.methodName,commandObj.argList,commandObj.argMap)
                        if tmpCacheKey is not None:
                            useCache = True
                            # cache is fresh
                            if tmpCacheKey in self.cacheMap and \
                               self.cacheMap[tmpCacheKey]['utime']+datetime.timedelta(seconds=timeRange) > datetime.datetime.utcnow():
                                tmpRet = self.cacheMap[tmpCacheKey]['value']
                                doExec = False
                    # exec
                    if doExec:
                        # get function
                        functionObj = getattr(self,commandObj.methodName)
                        # exec
                        tmpRet = functionObj(*commandObj.argList, **commandObj.argMap)
                    if isinstance(tmpRet,StatusCode):
                        # only status code was returned
                        retObj.statusCode = tmpRet
                    elif (isinstance(tmpRet, tuple) or isinstance(tmpRet, list)) \
                       and len(tmpRet) > 0 and isinstance(tmpRet[0],StatusCode):
                            retObj.statusCode = tmpRet[0]
                            # status code + return values
                            if len(tmpRet) > 1:
                                if retObj.statusCode == self.SC_SUCCEEDED:
                                    if len(tmpRet) == 2:
                                        retObj.returnValue = tmpRet[1]
                                    else:
                                        retObj.returnValue = tmpRet[1:]
                                else:
                                    if len(tmpRet) == 2:
                                        retObj.errorValue = tmpRet[1]
                                    else:
                                        retObj.errorValue = tmpRet[1:]
                    else:
                        retObj.statusCode = self.SC_SUCCEEDED
                        retObj.returnValue = tmpRet
                except Exception:
                    errtype,errvalue = sys.exc_info()[:2]
                    # failed
                    retObj.statusCode = self.SC_FATAL
                    retObj.errorValue = 'type=%s : %s.%s : %s' % \
                                        (errtype.__name__,className,
                                         commandObj.methodName,errvalue)
                # cache
                if useCache and doExec and retObj.statusCode == self.SC_SUCCEEDED:
                    self.cacheMap[tmpCacheKey] = {'utime':datetime.datetime.utcnow(),
                                                  'value':tmpRet}
            # return
            self.con.send(retObj)


# install SCs
installSC(CommandReceiveInterface)




###########################################################
#
# exceptions for IPC
#

# exception for temporary error
class JEDITemporaryError(Exception):
    pass


# exception for fatal error
class JEDIFatalError(Exception):
    pass


# exception for timeout error
class JEDITimeoutError(Exception):
    pass


# exception for no child error
class JEDINoChildError(Exception):
    pass



import re
import six
import random

from pandajedi.jedicore.MsgWrapper import MsgWrapper
from pandajedi.jedicore.SiteCandidate import SiteCandidate
from pandajedi.jedicore import Interaction
from pandajedi.jedicore import JediCoreUtils
from .JobBrokerBase import JobBrokerBase
from . import AtlasBrokerUtils
from pandajedi.jedirefine import RefinerUtils

# logger
from pandacommon.pandalogger.PandaLogger import PandaLogger
logger = PandaLogger().getLogger(__name__.split('.')[-1])


# brokerage for general purpose
class GenJobBroker (JobBrokerBase):

    # constructor
    def __init__(self,ddmIF,taskBufferIF):
        JobBrokerBase.__init__(self,ddmIF,taskBufferIF)


    # main
    def doBrokerage(self,taskSpec,cloudName,inputChunk,taskParamMap):
        # make logger
        tmpLog = MsgWrapper(logger,'<jediTaskID={0}>'.format(taskSpec.jediTaskID))
        tmpLog.debug('start')
        # return for failure
        retFatal    = self.SC_FATAL,inputChunk
        retTmpError = self.SC_FAILED,inputChunk
        # set cloud
        try:
            if not taskParamMap:
                taskParam = self.taskBufferIF.getTaskParamsWithID_JEDI(taskSpec.jediTaskID)
                taskParamMap = RefinerUtils.decodeJSON(taskParam)
            if not taskSpec.cloud and 'cloud' in taskParamMap:
                taskSpec.cloud = taskParamMap['cloud']
        except Exception:
            pass
        # get sites in the cloud
        site_preassigned = True
        if taskSpec.site not in ['', None]:
            tmpLog.debug('site={0} is pre-assigned'.format(taskSpec.site))
            if self.siteMapper.checkSite(taskSpec.site):
                scanSiteList = [taskSpec.site]
            else:
                scanSiteList = []
                for tmpSite in self.siteMapper.getCloud(taskSpec.cloud)['sites']:
                    if re.search(taskSpec.site, tmpSite):
                        scanSiteList.append(tmpSite)
                if not scanSiteList:
                    tmpLog.error('unknown site={}'.format(taskSpec.site))
                    taskSpec.setErrDiag(tmpLog.uploadLog(taskSpec.jediTaskID))
                    return retTmpError
        elif inputChunk.getPreassignedSite() is not None:
            scanSiteList = [inputChunk.getPreassignedSite()]
            tmpLog.debug('site={0} is pre-assigned in masterDS'.format(inputChunk.getPreassignedSite()))
        else:
            site_preassigned = False
            scanSiteList = self.siteMapper.getCloud(taskSpec.cloud)['sites']
            # remove NA
            if 'NA' in scanSiteList:
                scanSiteList.remove('NA')
            tmpLog.debug('cloud=%s has %s candidates' % (taskSpec.cloud, len(scanSiteList)))
        tmpLog.debug('initial {0} candidates'.format(len(scanSiteList)))
        ######################################
        # selection for status and PandaSite
        newScanSiteList = []
        for tmpSiteName in scanSiteList:
            tmpSiteSpec = self.siteMapper.getSite(tmpSiteName)
            # check site status
            if tmpSiteSpec.status != 'online' and not site_preassigned:
                tmpLog.debug('  skip %s due to status=%s' % (tmpSiteName, tmpSiteSpec.status))
                continue
            # check PandaSite
            if 'PandaSite' in taskParamMap and taskParamMap['PandaSite']:
                if tmpSiteSpec.pandasite != taskParamMap['PandaSite']:
                    tmpLog.debug('  skip %s due to wrong PandaSite=%s <> %s' % (tmpSiteName,
                                                                                tmpSiteSpec.pandasite,
                                                                                taskParamMap['PandaSite']))
                    continue
            newScanSiteList.append(tmpSiteName)
        scanSiteList = newScanSiteList
        tmpLog.debug('{0} candidates passed site status check'.format(len(scanSiteList)))
        if scanSiteList == []:
            tmpLog.error('no candidates')
            taskSpec.setErrDiag(tmpLog.uploadLog(taskSpec.jediTaskID))
            return retTmpError
        ######################################
        # selection for scratch disk
        minDiskCountS = taskSpec.getOutDiskSize() + taskSpec.getWorkDiskSize() + inputChunk.getMaxAtomSize()
        minDiskCountS = minDiskCountS // 1024 // 1024
        # size for direct IO sites
        if taskSpec.useLocalIO():
            minDiskCountR = minDiskCountS
        else:
            minDiskCountR = taskSpec.getOutDiskSize() + taskSpec.getWorkDiskSize()
            minDiskCountR = minDiskCountR // 1024 // 1024
        newScanSiteList = []
        for tmpSiteName in scanSiteList:
            tmpSiteSpec = self.siteMapper.getSite(tmpSiteName)
            # check at the site
            if tmpSiteSpec.maxwdir:
                if JediCoreUtils.use_direct_io_for_job(taskSpec, tmpSiteSpec, inputChunk):
                    minDiskCount = minDiskCountR
                else:
                    minDiskCount = minDiskCountS
                if minDiskCount > tmpSiteSpec.maxwdir:
                    tmpLog.debug('  skip {0} due to small scratch disk={1} < {2}'.format(tmpSiteName,
                                                                                         tmpSiteSpec.maxwdir,
                                                                                         minDiskCount))
                    continue
            newScanSiteList.append(tmpSiteName)
        scanSiteList = newScanSiteList
        tmpLog.debug('{0} candidates passed scratch disk check'.format(len(scanSiteList)))
        if scanSiteList == []:
            tmpLog.error('no candidates')
            taskSpec.setErrDiag(tmpLog.uploadLog(taskSpec.jediTaskID))
            return retTmpError
        ######################################
        # selection for available space in SE
        newScanSiteList = []
        for tmpSiteName in scanSiteList:
            # check at the site
            tmpSiteSpec = self.siteMapper.getSite(tmpSiteName)
            # free space must be >= 200GB
            diskThreshold = 200
            tmpSpaceSize = tmpSiteSpec.space
            if tmpSiteSpec.space and tmpSpaceSize < diskThreshold:
                tmpLog.debug('  skip {0} due to disk shortage in SE = {1} < {2}GB'.format(tmpSiteName,tmpSiteSpec.space,
                                                                                          diskThreshold))
                continue
            newScanSiteList.append(tmpSiteName)
        scanSiteList = newScanSiteList
        tmpLog.debug('{0} candidates passed SE space check'.format(len(scanSiteList)))
        if scanSiteList == []:
            tmpLog.error('no candidates')
            taskSpec.setErrDiag(tmpLog.uploadLog(taskSpec.jediTaskID))
            return retTmpError
        ######################################
        # selection for walltime
        minWalltime = taskSpec.walltime
        if minWalltime not in [0,None]:
            newScanSiteList = []
            for tmpSiteName in scanSiteList:
                tmpSiteSpec = self.siteMapper.getSite(tmpSiteName)
                # check at the site
                if tmpSiteSpec.maxtime != 0 and minWalltime > tmpSiteSpec.maxtime:
                    tmpLog.debug('  skip {0} due to short site walltime={1}(site upper limit) < {2}'.format(tmpSiteName,
                                                                                                            tmpSiteSpec.maxtime,
                                                                                                            minWalltime))
                    continue
                if tmpSiteSpec.mintime != 0 and minWalltime < tmpSiteSpec.mintime:
                    tmpLog.debug('  skip {0} due to short job walltime={1}(site lower limit) > {2}'.format(tmpSiteName,
                                                                                                           tmpSiteSpec.mintime,
                                                                                                           minWalltime))
                    continue
                newScanSiteList.append(tmpSiteName)
            scanSiteList = newScanSiteList
            tmpLog.debug('{0} candidates passed walltime check ={1}{2}'.format(len(scanSiteList),minWalltime,taskSpec.walltimeUnit))
            if scanSiteList == []:
                tmpLog.error('no candidates')
                taskSpec.setErrDiag(tmpLog.uploadLog(taskSpec.jediTaskID))
                return retTmpError
        ######################################
        # selection for memory
        origMinRamCount = inputChunk.getMaxRamCount()
        if not site_preassigned and origMinRamCount:
            newScanSiteList = []
            for tmpSiteName in scanSiteList:
                tmpSiteSpec = self.siteMapper.getSite(tmpSiteName)
                # job memory requirement
                if taskSpec.ramPerCore():
                    minRamCount = origMinRamCount * (tmpSiteSpec.coreCount if tmpSiteSpec.coreCount else 1)
                    minRamCount += (taskSpec.baseRamCount if taskSpec.baseRamCount else 0)
                else:
                    minRamCount = origMinRamCount
                # site max memory requirement
                site_maxmemory = tmpSiteSpec.maxrss if tmpSiteSpec.maxrss else 0
                # check at the site
                if site_maxmemory and minRamCount and minRamCount > site_maxmemory:
                    tmpMsg = '  skip site={0} due to site RAM shortage {1}(site upper limit) less than {2} '.format(tmpSiteName,
                                                                                                                    site_maxmemory,
                                                                                                                    minRamCount)
                    tmpLog.debug(tmpMsg)
                    continue
                # site min memory requirement
                site_minmemory = tmpSiteSpec.minrss if tmpSiteSpec.minrss else 0
                if site_minmemory and minRamCount and minRamCount < site_minmemory:
                    tmpMsg = '  skip site={0} due to job RAM shortage {1}(site lower limit) greater than {2} '.format(tmpSiteName,
                                                                                                                      site_minmemory,
                                                                                                                      minRamCount)
                    tmpLog.info(tmpMsg)
                    continue
                newScanSiteList.append(tmpSiteName)
            scanSiteList = newScanSiteList
            tmpLog.debug('{0} candidates passed memory check'.format(len(scanSiteList)))
            if scanSiteList == []:
                tmpLog.error('no candidates')
                taskSpec.setErrDiag(tmpLog.uploadLog(taskSpec.jediTaskID))
                return retTmpError
        ######################################
        # selection for nPilot
        nWNmap = self.taskBufferIF.getCurrentSiteData()
        newScanSiteList = []
        for tmpSiteName in scanSiteList:
            # check at the site
            nPilot = 0
            if tmpSiteName in nWNmap:
                nPilot = nWNmap[tmpSiteName]['getJob'] + nWNmap[tmpSiteName]['updateJob']
            if nPilot == 0 and taskSpec.prodSourceLabel not in ['test']:
                tmpLog.debug('  skip %s due to no pilot' % tmpSiteName)
                #continue
            newScanSiteList.append(tmpSiteName)
        scanSiteList = newScanSiteList
        tmpLog.debug('{0} candidates passed pilot activity check'.format(len(scanSiteList)))
        if scanSiteList == []:
            tmpLog.error('no candidates')
            taskSpec.setErrDiag(tmpLog.uploadLog(taskSpec.jediTaskID))
            return retTmpError
        ######################################
        # sites already used by task
        tmpSt,sitesUsedByTask = self.taskBufferIF.getSitesUsedByTask_JEDI(taskSpec.jediTaskID)
        if not tmpSt:
            tmpLog.error('failed to get sites which already used by task')
            taskSpec.setErrDiag(tmpLog.uploadLog(taskSpec.jediTaskID))
            return retTmpError
        ######################################
        # get list of available files
        availableFileMap = {}
        for datasetSpec in inputChunk.getDatasets():
            try:
                # get list of site to be scanned
                tmpLog.debug('getting the list of available files for {0}'.format(datasetSpec.datasetName))
                fileScanSiteList = []
                for tmpPseudoSiteName in scanSiteList:
                    tmpSiteSpec = self.siteMapper.getSite(tmpPseudoSiteName)
                    tmpSiteName = tmpSiteSpec.get_unified_name()
                    if tmpSiteName in fileScanSiteList:
                        continue
                    fileScanSiteList.append(tmpSiteName)
                # mapping between sites and input storage endpoints
                siteStorageEP = AtlasBrokerUtils.getSiteInputStorageEndpointMap(fileScanSiteList, self.siteMapper,
                                                                                taskSpec.prodSourceLabel, None)
                # disable file lookup for merge jobs
                if inputChunk.isMerging:
                    checkCompleteness = False
                else:
                    checkCompleteness = True
                if not datasetSpec.isMaster():
                    useCompleteOnly = True
                else:
                    useCompleteOnly = False
                # get available files per site/endpoint
                tmpAvFileMap = self.ddmIF.getAvailableFiles(datasetSpec,
                                                            siteStorageEP,
                                                            self.siteMapper,
                                                            check_completeness=checkCompleteness,
                                                            file_scan_in_container=False,
                                                            complete_only=useCompleteOnly)
                if tmpAvFileMap is None:
                    raise Interaction.JEDITemporaryError('ddmIF.getAvailableFiles failed')
                availableFileMap[datasetSpec.datasetName] = tmpAvFileMap
            except Exception as e:
                tmpLog.error('failed to get available files with {}'.format(e))
                taskSpec.setErrDiag(tmpLog.uploadLog(taskSpec.jediTaskID))
                return retTmpError
        ######################################
        # calculate weight
        tmpSt,jobStatPrioMap = self.taskBufferIF.getJobStatisticsByGlobalShare(taskSpec.vo)
        if not tmpSt:
            tmpLog.error('failed to get job statistics with priority')
            taskSpec.setErrDiag(tmpLog.uploadLog(taskSpec.jediTaskID))
            return retTmpError
        ######################################
        # final procedure
        tmpLog.debug('final {0} candidates'.format(len(scanSiteList)))
        weightMap = {}
        candidateSpecList = []
        preSiteCandidateSpec = None
        for tmpSiteName in scanSiteList:
            # get number of jobs in each job status. Using workQueueID=None to include non-JEDI jobs
            nRunning   = AtlasBrokerUtils.getNumJobs(jobStatPrioMap,tmpSiteName,'running',  None,None)
            nAssigned  = AtlasBrokerUtils.getNumJobs(jobStatPrioMap,tmpSiteName,'defined',  None,None)
            nActivated = AtlasBrokerUtils.getNumJobs(jobStatPrioMap,tmpSiteName,'activated',None,None)
            weight = float(nRunning + 1) / float(nActivated + nAssigned + 1) / float(nAssigned + 1)
            # make candidate
            siteCandidateSpec = SiteCandidate(tmpSiteName)
            # set weight
            siteCandidateSpec.weight = weight
            # files
            for tmpDatasetName, availableFiles in six.iteritems(availableFileMap):
                if tmpSiteName in availableFiles:
                    siteCandidateSpec.add_local_disk_files(availableFiles[tmpSiteName]['localdisk'])
            # append
            if tmpSiteName in sitesUsedByTask:
                candidateSpecList.append(siteCandidateSpec)
            else:
                if weight not in weightMap:
                    weightMap[weight] = []
                weightMap[weight].append(siteCandidateSpec)
        # limit the number of sites
        maxNumSites = 5
        weightList = list(weightMap.keys())
        weightList.sort()
        weightList.reverse()
        for weightVal in weightList:
            if len(candidateSpecList) >= maxNumSites:
                break
            sitesWithWeight = weightMap[weightVal]
            random.shuffle(sitesWithWeight)
            candidateSpecList += sitesWithWeight[:(maxNumSites-len(candidateSpecList))]
        # collect site names
        scanSiteList = []
        for siteCandidateSpec in candidateSpecList:
            scanSiteList.append(siteCandidateSpec.siteName)
        # append candidates
        newScanSiteList = []
        for siteCandidateSpec in candidateSpecList:
            # append
            inputChunk.addSiteCandidate(siteCandidateSpec)
            newScanSiteList.append(siteCandidateSpec.siteName)
            tmpLog.debug('  use {} with weight={} nFiles={}'.format(siteCandidateSpec.siteName,
                                                                    siteCandidateSpec.weight,
                                                                    len(siteCandidateSpec.localDiskFiles)))
        scanSiteList = newScanSiteList
        if scanSiteList == []:
            tmpLog.error('no candidates')
            taskSpec.setErrDiag(tmpLog.uploadLog(taskSpec.jediTaskID))
            return retTmpError
        # return
        tmpLog.debug('done')
        return self.SC_SUCCEEDED,inputChunk
